"""
Module contains models for MQTT communication with FIWARE's IoT-Agents.
"""

from aenum import Enum


class IoTAMQTTMessageType(str, Enum):
    """
    Options for mqtt message type
    """

    _init_ = "value __doc__"
    CMD = "cmd", "Command"
    CMDEXE = "cmdexe", "Command acknowledgement"
    MULTI = "multi", "Multi measurement"
    SINGLE = "single", "Single measurement"
    CONFIG = "configuration", "Configuration message"



"""Module contains the RDFParser that can create a Vocabulary object out of a
given ontology"""

import uuid
from enum import Enum
from typing import List, Tuple

import rdflib

from filip.models.base import LogLevel
from filip.semantics.ontology_parser.vocabulary_builder import VocabularyBuilder
from filip.semantics.vocabulary import (
    Source,
    IdType,
    Vocabulary,
    RestrictionType,
    ObjectProperty,
    DataProperty,
    Relation,
    TargetStatement,
    StatementType,
    DatatypeType,
    Datatype,
    Class,
    Individual,
)


specifier_base_iris = [
    "http://www.w3.org/2002/07/owl",
    "http://www.w3.org/1999/02/22-rdf-syntax-ns",
    "http://www.w3.org/XML/1998/namespace",
    "http://www.w3.org/2001/XMLSchema",
    "http://www.w3.org/2000/01/rdf-schema",
]
"""
Defines a set of base iris, that describe elements that belong to the 
description language not the ontology itself
"""


class Tags(str, Enum):
    """
    Collection of tags used as structures in ontologies, that were used more
    than once in the rdfparser code
    """

    rdf_type = ("http://www.w3.org/1999/02/22-rdf-syntax-ns#type",)
    owl_intersection = ("http://www.w3.org/2002/07/owl#intersectionOf",)
    owl_union = ("http://www.w3.org/2002/07/owl#unionOf",)
    owl_one_of = ("http://www.w3.org/2002/07/owl#oneOf",)
    owl_individual = ("http://www.w3.org/2002/07/owl#NamedIndividual",)
    owl_on_class = ("http://www.w3.org/2002/07/owl#onClass",)
    owl_on_data_range = "http://www.w3.org/2002/07/owl#onDataRange"


def get_iri_from_uriref(uriref: rdflib.URIRef) -> str:
    """Give an Uriref object, returns an iri

    Args:
        uriref: Object describing the iri

    Returns:
        str
    """
    return str(uriref)


def get_base_out_of_iri(iri: str) -> str:
    """Give an iri, returns an the ontology base name

    Args:
        iri

    Returns:
        str
    """
    if "#" in iri:
        index = iri.find("#")
        return iri[:index]
    else:
        # for example if uri looks like:
        # http://webprotege.stanford.edu/RDwpQ8vbi7HaApq8VoqJUXH
        index = iri.rfind("/")
        return iri[:index]


class RdfParser:
    """
    Class that parses a given source into a vocabulary.
    """

    def __init__(self):
        self.current_source = None
        """Current source which is parsed, used for Log entries"""
        self.current_class_iri = None
        """Iri of class which is currently parsed, used for Log entries"""

    def _add_logging_information(
        self, level: LogLevel, entity_type: IdType, entity_iri: str, msg: str
    ):
        """Add an entry to the parsing log

        Args:
            level (LogLevel): severe, warning or info
            entity_type (IdType)
            entity_iri (str)
            msg (str): Message to inform the user about the occurred issue

        Returns:
            None
        """
        if self.current_source is not None:
            self.current_source.add_parsing_log_entry(
                level, entity_type, entity_iri, msg
            )

    def parse_source_into_vocabulary(
        self, source: Source, vocabulary: Vocabulary
    ) -> bool:
        """Parse a Source into the given vocabulary
        Args:
            source (Source)
            vocabulary (Vocabulary)

        Returns:
            bool, True if success, False if Error occurred, as an invalid File
        """

        # if this is the predefined source don't parse it, just pretend it
        # was successful
        if source.predefined:
            return True

        voc_builder = VocabularyBuilder(vocabulary=vocabulary)
        g = rdflib.Graph()

        # format = rdflib.util.guess_format(source.source_path)
        voc_builder.add_source(source)
        voc_builder.set_current_source(source.id)

        g.parse(data=source.content, format="turtle")

        ontology_nodes = list(
            g.subjects(
                object=rdflib.term.URIRef("http://www.w3.org/2002/07/owl#Ontology"),
                predicate=rdflib.term.URIRef(Tags.rdf_type.value),
            )
        )

        # a source may have no ontology iri defined
        # if wanted on this place more info about the ontology can be extracted
        if len(ontology_nodes) > 0:
            source.ontology_iri = get_iri_from_uriref(ontology_nodes[0])

        self.current_source = source

        self._parse_to_vocabulary(g, voc_builder)

        return True

    def _is_object_defined_by_other_source(
        self, a: rdflib.term, graph: rdflib.Graph
    ) -> bool:
        """Test if the term is defined outside the current source

        Args:
            a (rdflib.term): Term to check
            graph (rdflib.graph): graph extracted from source

        Returns:
            bool
        """

        # if an object is defined by an other source it carries the predicate
        # ("isDefinedBy"). Then don't parse the object
        defined_tags = list(
            graph.objects(
                subject=a,
                predicate=rdflib.term.URIRef(
                    "http://www.w3.org/2000/01/rdf-schema#isDefinedBy"
                ),
            )
        )
        return len(defined_tags) > 0

    def _parse_to_vocabulary(self, graph: rdflib.Graph, voc_builder: VocabularyBuilder):
        """Parse an graph that was extracted from a TTL file into the vocabulary

        Args:
            graph (rdflib.Graph)
            voc_builder (VocabularyBuilder): Builder object to manipulate a
                vocabulary

        Returns:
            None
        """

        # OWLClasses
        for a in graph.subjects(
            object=rdflib.term.URIRef("http://www.w3.org/2002/07/owl#Class"),
            predicate=rdflib.term.URIRef(Tags.rdf_type.value),
        ):

            if isinstance(a, rdflib.term.BNode):
                pass
                # owl:Class can also occure in complex target statements of
                # relations as BNode, ignore it here
            else:

                # defined in other source -> ignore
                if self._is_object_defined_by_other_source(a, graph=graph):
                    continue

                iri, label, comment = self._extract_annotations(graph, a)
                c = Class(iri=iri, label=label, comment=comment)
                voc_builder.add_class(class_=c)

        # Class properties
        found_class_iris = set()
        for class_node in graph.subjects(
            predicate=rdflib.term.URIRef(
                "http://www.w3.org/2000/01/rdf-schema#subClassOf"
            )
        ):

            class_iri = get_iri_from_uriref(class_node)
            found_class_iris.add(class_iri)

        for class_iri in found_class_iris:
            # parent class / relation parsing
            for sub in graph.objects(
                subject=rdflib.term.URIRef(class_iri),
                predicate=rdflib.term.URIRef(
                    "http://www.w3.org/2000/01/rdf-schema#subClassOf"
                ),
            ):
                self.current_class_iri = class_iri  # used only for logging
                self._parse_subclass_term(
                    graph=graph, voc_builder=voc_builder, node=sub, class_iri=class_iri
                )

        # OWlObjectProperties
        for a in graph.subjects(
            object=rdflib.term.URIRef("http://www.w3.org/2002/07/owl#ObjectProperty"),
            predicate=rdflib.term.URIRef(Tags.rdf_type.value),
        ):

            if isinstance(a, rdflib.term.BNode):
                self._add_logging_information(
                    LogLevel.WARNING,
                    IdType.object_property,
                    "unknown",
                    "Found unparseable statement",
                )

            else:
                # defined in other source -> ignore
                if self._is_object_defined_by_other_source(a, graph):
                    continue

                iri, label, comment = self._extract_annotations(graph, a)

                obj_prop = ObjectProperty(iri=iri, label=label, comment=comment)
                voc_builder.add_object_property(obj_prop)
                # extract inverse properties, it can be multiple but only
                # URIRefs allowed no union/intersection
                for inverse_iri_node in graph.objects(
                    subject=a,
                    predicate=rdflib.term.URIRef(
                        "http://www.w3.org/2002/07/owl#inverseOf"
                    ),
                ):
                    if isinstance(inverse_iri_node, rdflib.term.BNode):
                        self._add_logging_information(
                            LogLevel.CRITICAL,
                            IdType.object_property,
                            iri,
                            "Complex inverseProperty statements aren't allowed",
                        )
                    else:
                        inverse_iri = get_iri_from_uriref(inverse_iri_node)
                        obj_prop.add_inverse_property_iri(inverse_iri)

        # OWlDataProperties
        for a in graph.subjects(
            object=rdflib.term.URIRef("http://www.w3.org/2002/07/owl#DatatypeProperty"),
            predicate=rdflib.term.URIRef(Tags.rdf_type.value),
        ):

            if isinstance(a, rdflib.term.BNode):
                self._add_logging_information(
                    LogLevel.WARNING,
                    IdType.data_property,
                    "unknown",
                    "Found unparseable statement",
                )

            else:
                # defined in other source -> ignore
                if self._is_object_defined_by_other_source(a, graph):
                    continue

                iri, label, comment = self._extract_annotations(graph, a)

                data_prop = DataProperty(iri=iri, label=label, comment=comment)
                voc_builder.add_data_property(data_prop)

        # OWLDataTypes
        # only the custom created datatype_catalogue are listed in the file,
        # the predefined are automatically added at the start
        # of post processing
        for a in graph.subjects(
            object=rdflib.term.URIRef("http://www.w3.org/2000/01/rdf-schema#Datatype"),
            predicate=rdflib.term.URIRef(Tags.rdf_type.value),
        ):

            if isinstance(a, rdflib.term.BNode):
                # self._add_logging_information(LogLevel.WARNING,
                #                              IdType.datatype, "unknown",
                #                              "Found unparseable statement")
                pass
                # e.g: :
                # customDataType4 rdf:type rdfs:Datatype ;
                # owl:equivalentClass [ rdf:type rdfs:Datatype ;....
                # the second Datatype triggers this if condition,
                # but we can ignore this statement

            else:
                # defined in other source -> ignore
                if self._is_object_defined_by_other_source(a, graph):
                    continue

                iri, label, comment = self._extract_annotations(graph, a)

                datatype = Datatype(iri=iri, label=label, comment=comment)
                voc_builder.add_datatype(datatype=datatype)

                # a datatype can be empty -> use string
                # a datatype can have multiple equivalent classes
                # (predefined types) -> ignore for now
                # a datatype can contain an enum of possible values ->
                # most interesting
                # under the predicate owl:equivalentClass is than a
                # list(first, rest, nil) under the pred.
                # oneOf with the values

                enum_values = []
                for equivalent_class in graph.objects(
                    subject=a,
                    predicate=rdflib.term.URIRef(
                        "http://www.w3.org/2002/07/owl#equivalentClass"
                    ),
                ):

                    if isinstance(equivalent_class, rdflib.term.URIRef):
                        # points to an other defined datatype, ignore
                        pass
                    else:
                        # is a bNode and points to owl:oneOf
                        enum_literals = self._extract_objects_out_of_single_combination(
                            graph,
                            equivalent_class,
                            accept_and=False,
                            accept_or=False,
                            accept_one_of=True,
                        )
                        for literal in enum_literals:
                            enum_values.append(str(literal))
                datatype.enum_values = enum_values
                if len(enum_values) > 0:
                    datatype.type = DatatypeType.enum
                else:
                    datatype.type = DatatypeType.string

        # OWLIndividuals

        for a in graph.subjects(
            object=rdflib.term.URIRef(Tags.owl_individual.value),
            predicate=rdflib.term.URIRef(Tags.rdf_type.value),
        ):

            if isinstance(a, rdflib.term.BNode):
                self._add_logging_information(
                    LogLevel.WARNING,
                    IdType.individual,
                    "unknown",
                    "Found unparseable statement",
                )

            else:
                # defined in other source -> ignore
                if self._is_object_defined_by_other_source(a, graph):
                    continue

                iri, label, comment = self._extract_annotations(graph, a)
                objects = graph.objects(
                    subject=a, predicate=rdflib.term.URIRef(Tags.rdf_type.value)
                )
                # superclasses = types
                types = []
                for object in objects:
                    if not object == rdflib.term.URIRef(Tags.owl_individual.value):
                        types.extend(
                            self._extract_objects_out_of_layered_combination(
                                graph, object, True, False
                            )
                        )

                individual = Individual(iri=iri, label=label, comment=comment)
                for type in types:
                    individual.parent_class_iris.append(get_iri_from_uriref(type))
                voc_builder.add_individual(individual=individual)

        # As seen for example in the bricks ontology an individual can be
        # declared with :individual1 rdf:type :Class1
        # this type of declaration is hard to completly detect
        # we need to see that the object is a class iri and not a specifier iri.
        # as we may not have loaded all dependencies we can not simply look it
        # up in vocabulary
        # -> getbase uri of statement and filter all known specifier uris
        for sub in graph.subjects(predicate=rdflib.term.URIRef(Tags.rdf_type.value)):
            for obj in graph.objects(
                subject=sub, predicate=rdflib.term.URIRef(Tags.rdf_type.value)
            ):

                if isinstance(obj, rdflib.term.BNode):
                    continue
                obj_iri = get_iri_from_uriref(obj)

                obj_base_iri = get_base_out_of_iri(iri=obj_iri)
                if obj_base_iri not in specifier_base_iris:
                    iri, label, comment = self._extract_annotations(graph, sub)
                    if not voc_builder.entity_is_known(iri):
                        iri, label, comment = self._extract_annotations(graph, sub)
                        individual = Individual(iri=iri, label=label, comment=comment)
                        individual.parent_class_iris.append(obj_iri)
                        voc_builder.add_individual(individual)

    def _extract_annotations(
        self, graph: rdflib.Graph, node: rdflib.term.URIRef
    ) -> Tuple[str, str, str]:
        """Extract out of a node term the owl annotations (iri, label, comment)

        Args:
            graph (rdflib.graph): Graph describing ontology
            node (rdflib.term.URIRef): Entity node

        Returns:
            [str,str,str]: [iri, label, comment]
        """
        iri = str(node)
        label = graph.label(node).title()
        comment = graph.comment(node).title()

        return iri, label, comment

    def _parse_subclass_term(
        self,
        graph: rdflib.Graph,
        voc_builder: VocabularyBuilder,
        node: rdflib.term,
        class_iri: str,
    ):
        """Parse a subclass term of the given node and class_iri

        Args:
            graph (rdflib.graph): Graph describing ontology
            vocabulary (Vocabulary): Vocabualry to parse into
            node (rdflib.term)
            class_iri (str)

        Returns:
            None
        """

        # class could have been only defined in other source, than no class
        # is defined, but as we have found a relation for a class, the class
        # needs to exist
        if class_iri not in voc_builder.vocabulary.classes:
            voc_builder.add_class(class_=Class(iri=class_iri))

        # node can be 1 of 3 things:
        #   - a parentclass statment -> UriRef
        #   - a relation statment -> BNode
        #   - an intersection of parentclasses ,
        #   relations and intersections -> BNode
        if isinstance(node, rdflib.term.BNode):
            # sub has no IRI and is therefore a relation

            # extract the subpredicates and subobjects as statments
            # if node is a relation:
            #      in total there should be 3-4 statments:
            #      rdf:type pointing to owl:Restriction
            #      owl:onProperty pointing to a data or object property
            #      1-2 staments which values are exepted, this can point to an
            #      URIRef or BNode

            # if node is a intersection:
            #      it has the predicate owl:intersectionOf
            #      and a set of objects

            predicates = []
            objects = []
            for p in graph.predicates(subject=node):
                predicates.append(p)
            for o in graph.objects(subject=node):
                objects.append(o)

            # Combination of statements
            if rdflib.term.URIRef(Tags.owl_intersection.value) in predicates:
                objects = self._extract_objects_out_of_single_combination(
                    graph, node, True, False
                )
                for object in objects:
                    self._parse_subclass_term(
                        graph=graph,
                        voc_builder=voc_builder,
                        node=object,
                        class_iri=class_iri,
                    )

            elif rdflib.term.URIRef(Tags.owl_union.value) in predicates:
                self._add_logging_information(
                    LogLevel.CRITICAL,
                    IdType.class_,
                    class_iri,
                    "Relation statements combined with or",
                )

            elif rdflib.term.URIRef(Tags.owl_one_of.value) in predicates:
                self._add_logging_information(
                    LogLevel.CRITICAL,
                    IdType.class_,
                    class_iri,
                    "Relation statements combined with oneOf",
                )

            # Relation statement
            else:

                additional_statements = {}
                rdf_type = ""
                owl_on_property = ""

                for i in range(len(predicates)):
                    if predicates[i] == rdflib.term.URIRef(Tags.rdf_type.value):
                        rdf_type = get_iri_from_uriref(objects[i])
                    elif predicates[i] == rdflib.term.URIRef(
                        "http://www.w3.org/2002/07/owl#onProperty"
                    ):
                        owl_on_property = get_iri_from_uriref(objects[i])
                    else:
                        additional_statements[get_iri_from_uriref(predicates[i])] = (
                            objects[i]
                        )

                relation_is_ok = True
                if not rdf_type == "http://www.w3.org/2002/07/owl#Restriction":
                    self._add_logging_information(
                        LogLevel.CRITICAL,
                        IdType.class_,
                        class_iri,
                        "Class has an unknown subClass statement",
                    )
                    relation_is_ok = False

                if owl_on_property == "":
                    self._add_logging_information(
                        LogLevel.CRITICAL,
                        IdType.class_,
                        class_iri,
                        "Class has a relation without a property",
                    )
                    relation_is_ok = False

                # object or data relation?
                if relation_is_ok:
                    relation = None
                    id = uuid.uuid4().hex
                    # this id can and should be random. a class_iri can have a
                    # property_iri multiple times, to assign always the same id
                    # for the same relation is not worth the trouble

                    relation = Relation(property_iri=owl_on_property, id=id)
                    voc_builder.add_relation_for_class(class_iri, relation)

                    # go through the additional statement to figure out the
                    # targetIRI and the restrictionType/cardinality
                    self._parse_relation_type(graph, relation, additional_statements)

        # parent-class statement or empty list element
        else:
            # owlThing is the root object, but it is not declared as a class
            # in the file to prevent None pointer when looking up parents,
            # a class that has a parent owlThing simply has no parents
            if (
                not get_iri_from_uriref(node)
                == "http://www.w3.org/1999/02/22-rdf-syntax-ns#nil"
            ):
                # ignore empty lists
                if (
                    not get_iri_from_uriref(node)
                    == "http://www.w3.org/2002/07/owl#Thing"
                ):
                    voc_builder.vocabulary.get_class_by_iri(
                        class_iri
                    ).parent_class_iris.append(get_iri_from_uriref(node))

    def _parse_relation_type(
        self, graph: rdflib.Graph, relation: Relation, statements: {}
    ):
        """
        Parse the relation type and depending on the result the
        cardinality or value of relation

        Args:
            graph: underlying ontology graph
            relation: Relation object into which the information are saved
            statements: Ontology statements concerning the relation

        Returns:
            None
        """
        treated_statements = []
        for statement in statements:
            if statement == "http://www.w3.org/2002/07/owl#someValuesFrom":
                relation.restriction_type = RestrictionType.some
                self._parse_relation_values(graph, relation, statements[statement])
            elif statement == "http://www.w3.org/2002/07/owl#allValuesFrom":
                relation.restriction_type = RestrictionType.only
                self._parse_relation_values(graph, relation, statements[statement])
            elif statement == "http://www.w3.org/2002/07/owl#hasValue":
                relation.restriction_type = RestrictionType.value
                # has Value can only point to a single value
                self._parse_has_value(graph, relation, statements[statement])
            elif statement == "http://www.w3.org/2002/07/owl#maxCardinality":
                relation.restriction_type = RestrictionType.max
                self._parse_cardinality(
                    graph, relation, statement, statements, treated_statements
                )
            elif statement == "http://www.w3.org/2002/07/owl#minCardinality":
                relation.restriction_type = RestrictionType.min
                self._parse_cardinality(
                    graph, relation, statement, statements, treated_statements
                )
            elif statement == "http://www.w3.org/2002/07/owl#cardinality":
                relation.restriction_type = RestrictionType.exactly
                self._parse_cardinality(
                    graph, relation, statement, statements, treated_statements
                )
            elif statement == "http://www.w3.org/2002/07/owl#maxQualifiedCardinality":
                relation.restriction_type = RestrictionType.max
                self._parse_cardinality(
                    graph, relation, statement, statements, treated_statements
                )
            elif statement == "http://www.w3.org/2002/07/owl#minQualifiedCardinality":
                relation.restriction_type = RestrictionType.min
                self._parse_cardinality(
                    graph, relation, statement, statements, treated_statements
                )
            elif statement == "http://www.w3.org/2002/07/owl#qualifiedCardinality":
                relation.restriction_type = RestrictionType.exactly
                self._parse_cardinality(
                    graph, relation, statement, statements, treated_statements
                )

            treated_statements.append(statement)

        for statement in statements:
            if statement not in treated_statements:
                self._add_logging_information(
                    LogLevel.CRITICAL,
                    IdType.class_,
                    self.current_class_iri,
                    "Relation with property {} has an untreated restriction "
                    "{}".format(relation.property_iri, statement),
                )

    def _parse_cardinality(
        self,
        graph: rdflib.Graph,
        relation: Relation,
        statement,
        statements,
        treated_statements,
    ):
        """Parse the cardinality of a relation

        Args:
            graph: underlying ontology graph
            relation: Relation object into which the information are saved
            statement: The statement that is actively treated
            statements: Ontology statements concerning the relation
            treated_statements: Statements that were already treated

        Returns:
            None
        """
        if Tags.owl_on_class.value in statements:
            relation.restriction_cardinality = str(statements[statement])
            target = statements[Tags.owl_on_class.value]
            self._parse_relation_values(graph, relation, target)
            treated_statements.append(Tags.owl_on_class.value)
        elif Tags.owl_on_data_range.value in statements:
            relation.restriction_cardinality = str(statements[statement])
            target = statements[Tags.owl_on_data_range.value]
            self._parse_relation_values(graph, relation, target)
            treated_statements.append(Tags.owl_on_data_range.value)
        else:
            # has From:
            # in File: owl:maxCardinality "1"^^xsd:nonNegativeInteger
            # e.g.: {'http://www.w3.org/2002/07/owl#maxCardinality':
            #        rdflib.term.Literal('1', datatype=
            #        rdflib.term.URIRef('
            #        http://www.w3.org/2001/XMLSchema#nonNegativeInteger'))}

            # in this case the file does not state a datarange that is allowed.
            # Therefore the target gets set to the universal string

            relation.restriction_cardinality = statements[statement].value
            datatype = "http://www.w3.org/2001/XMLSchema#string"
            target_statement = TargetStatement(
                type=StatementType.LEAF, target_iri=datatype
            )
            relation.target_statement = target_statement

    def _parse_has_value(
        self, graph: rdflib.Graph, relation: Relation, node: rdflib.term
    ):
        """Parse the value of a relation

        Args:
           graph: underlying ontology graph
           relation: Relation object into which the information are saved
           node: (complex) Graph node containing the value

        Returns:
           None
        """
        self._parse_relation_values(graph, relation, node)
        # for hasValue only a target-statement that is a leaf is allowed
        if not relation.target_statement.type == StatementType.LEAF:
            self._add_logging_information(
                LogLevel.CRITICAL,
                IdType.class_,
                self.current_class_iri,
                f"In hasValue relation with property {relation.property_iri} "
                f"target is a complex expression",
            )

    def _parse_relation_values(
        self, graph: rdflib.Graph, relation: Relation, node: rdflib.term
    ):
        """
        Parse the value of a relation out of a node that can be complex;
        consisting out of a combination of multiple other nodes

        Args:
           graph: underlying ontology graph
           relation: Relation object into which the information are saved
           node: (complex) Graph node containing the value

        Returns:
           None
        """
        target_statement = TargetStatement()
        relation.target_statement = target_statement

        queue = [(node, target_statement)]
        while not len(queue) == 0:
            current_term, current_statement = queue.pop(0)
            if isinstance(current_term, rdflib.URIRef):
                target_iri = get_iri_from_uriref(current_term)

                current_statement.set_target(target_iri=target_iri)
            else:
                if rdflib.term.URIRef(Tags.owl_intersection.value) in graph.predicates(
                    subject=current_term
                ):

                    current_statement.type = StatementType.AND
                elif rdflib.term.URIRef(Tags.owl_union.value) in graph.predicates(
                    subject=current_term
                ):

                    current_statement.type = StatementType.OR
                else:
                    current_statement.set_target(
                        target_iri="Target statement has no iri",
                        target_data_value=str(current_term),
                    )

                    continue

                child_nodes = self._extract_objects_out_of_single_combination(
                    graph, current_term, True, True
                )
                for child_node in child_nodes:
                    new_statement = TargetStatement()
                    current_statement.target_statements.append(new_statement)
                    queue.append((child_node, new_statement))

    # an intersection/union is a basic list, it consists out of a chain of
    # bnode, where each bnode has the "first"and "rest" predicate, first
    # contains our object, rest is a pointer to the next part of the chain.
    # the list is over if rest points to "NIL"
    # this methode extracts all objects of a single layered intersection,
    # if the intersection contains further intersections these are contained in
    # the result list as BNode
    def _extract_objects_out_of_single_combination(
        self,
        graph: rdflib.Graph,
        node: rdflib.term.BNode,
        accept_and: bool,
        accept_or: bool,
        accept_one_of: bool = False,
    ):
        """
        An intersection/union is a basic list, it consits out of a chain of
        bnode,where each bnode has the "first"and "rest" predicate,
        first contains our object, rest is a pointer to the next part of the
        chain. The list is over if rest points to "NIL"
        This methode extracts all objects of a single layered intersection,
        if the intersection contains further intersections these are contained
        in the result list as BNode

        Args:
            graph: underlying ontology graph
            node: (complex) Graph node containing the value
            accept_or (bool): true, if combinations with "or" are allowed to be
                parsed
            accept_and (bool): true, if combinations with "and" are allowed
                to be parsed
            accept_one_of (bool): true, if ne_of statements are allowed
                to be parsed

        Returns:
           None
        """
        predicates = list(graph.predicates(subject=node))

        # the passed startnode needs to contain an intersection or a union
        # both at the same time should not be possible
        start_node = None
        if rdflib.term.URIRef(Tags.owl_intersection.value) in predicates:
            if accept_and:
                start_node = next(
                    graph.objects(
                        subject=node,
                        predicate=rdflib.term.URIRef(Tags.owl_intersection.value),
                    )
                )
        elif rdflib.term.URIRef(Tags.owl_union.value) in predicates:
            if accept_or:
                start_node = next(
                    graph.objects(
                        subject=node, predicate=rdflib.term.URIRef(Tags.owl_union.value)
                    )
                )
        elif rdflib.term.URIRef(Tags.owl_one_of.value) in predicates:
            if accept_one_of:
                start_node = next(
                    graph.objects(
                        subject=node,
                        predicate=rdflib.term.URIRef(Tags.owl_one_of.value),
                    )
                )
        else:
            self._add_logging_information(
                LogLevel.CRITICAL,
                IdType.class_,
                self.current_class_iri,
                f"Intern Error - invalid {node} passed to list extraction",
            )

        result = []
        rest = start_node
        if start_node is None:
            return []

        while not rest == rdflib.term.URIRef(
            "http://www.w3.org/1999/02/22-rdf-syntax-ns#nil"
        ):

            first = next(
                graph.objects(
                    predicate=rdflib.term.URIRef(
                        "http://www.w3.org/1999/02/22-rdf-syntax-ns#first"
                    ),
                    subject=rest,
                )
            )
            result.append(first)
            rest = next(
                graph.objects(
                    predicate=rdflib.term.URIRef(
                        "http://www.w3.org/1999/02/22-rdf-syntax-ns#rest"
                    ),
                    subject=rest,
                )
            )

        return result

    def _extract_objects_out_of_layered_combination(
        self,
        graph: rdflib.Graph,
        node: rdflib.term.BNode,
        accept_and: bool,
        accept_or: bool,
    ) -> List[rdflib.term.URIRef]:
        """Extract all nodes out of a complex combination

        Args:
            graph: underlying ontology graph
            node: (complex) Graph node containing the complex combination
            accept_or (bool): true, if combinations with "or" are allowed to be
                parsed
            accept_and (bool): true, if combinations with "and" are allowed
                to be parsed

        Returns:
           List[rdflib.term.URIRef], list of terms out of combination
        """
        result = []
        queue = [node]

        while len(queue) > 0:
            node = queue.pop()
            if isinstance(node, rdflib.term.URIRef):
                result.append(node)
            else:
                queue.extend(
                    self._extract_objects_out_of_single_combination(
                        graph, node, accept_and, accept_or
                    )
                )
        return result



"""
Helper functions to prohibit boiler plate code
"""

import logging
import re
import warnings
from aenum import Enum
from typing import Dict, Any, List
from pydantic import AnyHttpUrl, validate_call
from pydantic_core import PydanticCustomError
from filip.custom_types import AnyMqttUrl
from pyjexl.jexl import JEXL
from pyjexl.parser import Transform
from pyjexl.exceptions import ParseError

logger = logging.getLogger(name=__name__)


class FiwareRegex(str, Enum):
    """
    Collection of Regex expression used to check if the value of a Pydantic
    field, can be used in the related Fiware field. The regexes here are primarily
    defined based on the identifiers syntax restriction:
    https://fiware-orion.readthedocs.io/en/stable/orion-api.html#identifiers-syntax-restrictions
    """

    _init_ = "value __doc__"
    #  Identifiers syntax restriction
    standard = (
        r"(^((?![?&#/\"' ])[\x00-\x7F])*$)",
        "Prevents any string that contains at least one of the "
        "symbols: ? & # / ' \" or a whitespace",
    )
    string_protect = (
        r"(?!^id$)(?!^type$)(?!^geo:json$)(^((?![?&#/\"' ])[\x00-\x7F])*$)",
        "Prevents any string that contains at least one of "
        "the symbols: ? & # / ' \" or a whitespace."
        "AND the strings: id, type, geo:json",
    )
    attribute_name = (
        r"(?!^id$)(?!^type$)(?!^geo:json$)(^((?![\"'<>()=; §&/#?])[\x00-\x7F])*$)",
        "Prevents any string that contains at least one of the "
        "symbols: ( ) < > \" ' = ; § & / # ?",
    )
    attribute_value = (
        r"(^((?![\"'<>()=;]).)*$)",
        "Prevents any string that contains at least one of the "
        "symbols: ( ) < > \" ' = ; ",
    )


@validate_call
def validate_http_url(url: AnyHttpUrl) -> str:
    """
    Function checks whether the host has "http" added in case of http as
    protocol.

    Args:
        url (AnyHttpUrl): the url for the host / port

    Returns:
        validated url
    """
    url = str(url) if url else url
    if url[-1] != "/":
        # add trailing slash
        url = f"{url}/"
    return url


@validate_call
def validate_mqtt_url(url: AnyMqttUrl) -> str:
    """
    Function that checks whether a url is valid mqtt endpoint

    Args:
        url: the url for the target endpoint

    Returns:
       validated url
    """
    return str(url) if url else url


def validate_escape_character_free(value: Any) -> Any:
    """
    Function that checks whether a value contains a string part that starts
    or end with ' or ".
    the function iterates to break down each complex data-structure to its
    fundamental string parts.
    Each value of a list is examined
    Of dictionaries each value is examined, keys are skipped, as they are ok
    for Fiware

    Args:
        value: the string to check

    Returns:
       validated string
    """

    if not isinstance(value, List):
        values = [value]
    else:
        values = value

    for value in values:
        if isinstance(value, Dict):
            for key, dict_value in value.items():
                validate_escape_character_free(dict_value)
                # it seems Fiware has no problem if the keys contain ' or "
                # validate_escape_character_free(key)
        elif isinstance(value, List):
            for inner_list in value:
                validate_escape_character_free(inner_list)
        else:
            # if a value here is not a string, it will also not contain ' or "
            value = str(value)
            if '"' == value[-1:] or '"' == value[0:1]:
                raise ValueError(f"The value {value} contains " f'the forbidden char "')
            if "'" == value[-1:] or "'" == value[0:1]:
                raise ValueError(f"The value {value} contains " f"the forbidden char '")
    return values


def match_regex(value: str, pattern: str):
    regex = re.compile(pattern)
    if not regex.match(value):
        raise PydanticCustomError(
            "string_pattern_mismatch",
            "String should match pattern '{pattern}', [type='{error_type}', input_value='{value}']",
            {
                "pattern": pattern,
                "error_type": "string_pattern_mismatch",
                "value": value,
            },
        )
    return value


def ignore_none_input(func):
    def wrapper(arg):
        if arg is None:
            return arg
        return func(arg)

    return wrapper


def validate_fiware_standard_regex(vale: str):
    return match_regex(vale, FiwareRegex.standard.value)


def validate_fiware_string_protect_regex(vale: str):
    return match_regex(vale, FiwareRegex.string_protect.value)


def validate_fiware_attribute_value_regex(vale: str):
    return match_regex(vale, FiwareRegex.attribute_value.value)


def validate_fiware_attribute_name_regex(vale: str):
    return match_regex(vale, FiwareRegex.attribute_name.value)


@ignore_none_input
def validate_mqtt_topic(topic: str):
    return match_regex(topic, r"^((?![\'\"#+,])[\x00-\x7F])*$")


@ignore_none_input
def validate_fiware_datatype_standard(_type):
    from filip.models.base import DataType

    if isinstance(_type, DataType):
        return _type
    elif isinstance(_type, str):
        return validate_fiware_standard_regex(_type)
    else:
        raise TypeError(f"Invalid type {type(_type)}")


@ignore_none_input
def validate_fiware_datatype_string_protect(_type):
    from filip.models.base import DataType

    if isinstance(_type, DataType):
        return _type
    elif isinstance(_type, str):
        return validate_fiware_string_protect_regex(_type)
    else:
        raise TypeError(f"Invalid type {type(_type)}")


@ignore_none_input
def validate_fiware_service_path(service_path):
    return match_regex(service_path, r"^((\/\w*)|(\/\#))*(\,((\/\w*)|(\/\#)))*$")


@ignore_none_input
def validate_fiware_service(service):
    return match_regex(service, r"\w*$")


jexl_transformation_functions = {
    "jsonparse": "(str) => JSON.parse(str)",
    "jsonstringify": "(obj) => JSON.stringify(obj)",
    "indexOf": "(val, char) => String(val).indexOf(char)",
    "length": "(val) => String(val).length",
    "trim": "(val) => String(val).trim()",
    "substr": "(val, int1, int2) => String(val).substr(int1, int2)",
    "addreduce": "(arr) => arr.reduce((i, v) => i + v)",
    "lengtharray": "(arr) => len(arr)",
    "typeof": "(val) => typeof val",
    "isarray": "(arr) => Array.isArray(arr)",
    "isnan": "(val) => isNaN(val)",
    "parseint": "(val) => parseInt(val)",
    "parsefloat": "(val) => parseFloat(val)",
    "toisodate": "(val) => new Date(val).toISOString()",
    "timeoffset": "(isostr) => new Date(isostr).getTimezoneOffset()",
    "tostring": "(val) => str(val)",
    "urlencode": "(val) => encodeURI(val)",
    "urldecode": "(val) => decodeURI(val)",
    "replacestr": "(str, from, to) => str.replace(from, to)",
    "replaceregexp": "(str, reg, to) => str.replace(reg, to)",
    "replaceallstr": "(str, from, to) => str.replace(from, to)",
    "replaceallregexp": "(str, reg, to) => str.replace(reg, to)",
    "split": "(str, ch) => str.split(ch)",
    "mapper": "(val, values, choices) => choices[values.index(val)]",
    "thmapper": "(val, values, choices) => choices[next((i for i, v in enumerate(values) if val <= v), None)]",
    "bitwisemask": "(i, mask, op, shf) => ((int(i) & mask) if op == '&' else ((int(i) | mask) if op == '|' else ((int(i) ^ mask) if op == '^' else int(i))) >> shf)",
    "slice": "(arr, init, end) => arr[init:end]",
    "addset": "(arr, x) => list(set(arr).add(x))",
    "removeset": "(arr, x) => list(set(arr).remove(x))",
    "touppercase": "(val) => str(val).upper()",
    "tolowercase": "(val) => str(val).lower()",
}


def validate_jexl_expression(expression, attribute_name, device_id):
    try:
        jexl_expression = JEXL().parse(expression)
        if isinstance(jexl_expression, Transform):
            if jexl_expression.name not in jexl_transformation_functions.keys():
                warnings.warn(f"{jexl_expression.name} might not supported")
    except ParseError:
        msg = f"Invalid JEXL expression '{expression}' inside the attribute '{attribute_name}' of Device '{device_id}'."
        if "|" in expression:
            msg += " If the expression contains the transform operator '|' you need to remove the spaces around it."
        raise ParseError(msg)
    return expression


def validate_expression_language(cls, expressionLanguage):
    if expressionLanguage == "legacy":
        warnings.warn(
            f"Using 'LEGACY' expression language inside {cls.__name__} is "
            f"deprecated. Use 'JEXL' instead."
        )
    elif expressionLanguage is None:
        expressionLanguage = "jexl"
    return expressionLanguage



"""Module providing an interface to manipulate the sources of a vocabulary,
and to ability to export it to models"""

import copy
import io
import keyword
import os
from datetime import datetime
from string import ascii_letters, digits
from typing import List, Optional, Dict, Tuple, Set

import pathlib
import requests
import wget

from filip.semantics.ontology_parser.post_processer import PostProcessor
from filip.semantics.ontology_parser.rdfparser import RdfParser
from filip.semantics.vocabulary import (
    LabelSummary,
    Vocabulary,
    Source,
    Entity,
    RestrictionType,
    Class,
    ParsingError,
    CombinedRelation,
    DataFieldType,
    DependencyStatement,
    VocabularySettings,
)

# Blacklist containing all labels that are forbidden for entities to have
label_blacklist = list(keyword.kwlist)
label_blacklist.extend(["referencedBy", "deviceSettings"])
label_blacklist.extend(
    [
        "references",
        "device_settings",
        "header",
        "old_state",
        "",
        "semantic_manager",
        "delete",
        "metadata",
    ]
)
label_blacklist.extend(["id", "type", "class"])
label_blacklist.extend(
    [
        "str",
        "int",
        "float",
        "complex",
        "list",
        "tuple",
        "range",
        "dict",
        "list",
        "set",
        "frozenset",
        "bool",
        "bytes",
        "bytearray",
        "memoryview",
    ]
)

# Whitelist containing all chars that an entity label can consist of
label_char_whitelist = ascii_letters + digits + "_"


class VocabularyConfigurator:
    """
    Class that provides static interfaces to manipulate the sources of a
    vocabulary, validate and save it.
    """

    @classmethod
    def create_vocabulary(
        cls, settings: VocabularySettings = VocabularySettings()
    ) -> Vocabulary:
        """
        Create a new blank vocabulary with given settings

        Args:
            settings: VocabularySettings

        Returns:
            Vocabulary
        """

        return Vocabulary(settings=settings)

    @classmethod
    def delete_source_from_vocabulary(
        cls, vocabulary: Vocabulary, source_id: str
    ) -> Vocabulary:
        """
        Delete a source from the vocabulary

        Args:
            vocabulary (Vocabulary): Vocabulary from which the source should
            be removed
            source_id (str): Id of source to remove

        Raises:
            ValueError:  If no source with given Id exists in Vocabulary

        Returns:
            New Vocabulary without the given source
        """
        new_vocabulary = Vocabulary(settings=copy.copy(vocabulary.settings))
        parser = RdfParser()
        found = False
        for source in vocabulary.sources.values():
            if not source_id == source.id:
                parser.parse_source_into_vocabulary(
                    source=copy.deepcopy(source), vocabulary=new_vocabulary
                )
            else:
                found = True

        PostProcessor.post_process_vocabulary(
            vocabulary=new_vocabulary, old_vocabulary=vocabulary
        )

        if not found:
            raise ValueError(f"Source with source_id {source_id} not in vocabulary")

        PostProcessor.transfer_settings(
            new_vocabulary=new_vocabulary, old_vocabulary=vocabulary
        )

        return new_vocabulary

    @classmethod
    def add_ontology_to_vocabulary_as_link(
        cls, vocabulary: Vocabulary, link: str, source_name: Optional[str] = None
    ) -> Vocabulary:
        """
        Add a source to the vocabulary with via a weblink. Source name will
        be extracted from link, if no name is given

        Args:
            vocabulary (Vocabulary): Vocabulary to which the source should
            be added
            link (str): Weblink to the source
            source_name (Optional[str]): Name for the source

        Raises:
            ParsingException:  If the given source was not valid and could not
            be parsed

        Returns:
            New Vocabulary with the given source added to it
        """

        downloaded_obj = requests.get(link)
        file_bytes = io.BytesIO(downloaded_obj.content)
        if source_name is None:
            source_name = wget.filename_from_url(link)

        file_str = io.TextIOWrapper(file_bytes, encoding="utf-8").read()

        return cls.add_ontology_to_vocabulary_as_string(
            vocabulary=vocabulary, source_name=source_name, source_content=file_str
        )

    @classmethod
    def add_ontology_to_vocabulary_as_file(
        cls,
        vocabulary: Vocabulary,
        path_to_file: str,
        source_name: Optional[str] = None,
    ) -> Vocabulary:
        """
        Add a source to the vocabulary with via a file path. Source name will
        be extracted from path, if no name is given

        Args:
            vocabulary (Vocabulary): Vocabulary to which the source should
            be added
            path_to_file (str): Path to the source file
            source_name (Optional[str]): Name for the source

        Raises:
            ParsingException:  If the given source was not valid and could not
            be parsed

        Returns:
            New Vocabulary with the given source added to it
        """

        with open(path_to_file, "r") as file:
            data = file.read()

        if source_name is None:
            source_name = os.path.basename(path_to_file).split(".")[0]

        source = Source(source_name=source_name, content=data, timestamp=datetime.now())

        return VocabularyConfigurator._parse_sources_into_vocabulary(
            vocabulary=vocabulary, sources=[source]
        )

    @classmethod
    def add_ontology_to_vocabulary_as_string(
        cls, vocabulary: Vocabulary, source_name: str, source_content: str
    ) -> Vocabulary:
        """
        Add a source to the vocabulary by giving the source content as string.
        Source name needs to be given

        Args:
            vocabulary (Vocabulary): Vocabulary to which the source should
            be added
            source_content (str): Content of source
            source_name (str): Name for the source

        Raises:
            ParsingException:  If the given source was not valid and could not
            be parsed

        Returns:
            New Vocabulary with the given source added to it
        """
        source = Source(
            source_name=source_name, content=source_content, timestamp=datetime.now()
        )

        return VocabularyConfigurator._parse_sources_into_vocabulary(
            vocabulary=vocabulary, sources=[source]
        )

    @classmethod
    def _parse_sources_into_vocabulary(
        cls, vocabulary: Vocabulary, sources: List[Source]
    ) -> Vocabulary:
        """
        Parse the given source objects into the vocabulary

        Args:
            vocabulary (Vocabulary): Vocabulary to which the source should
            be added
            sources (List[Source]): Source objects to be added

        Raises:
            ParsingException:  If the given source was not valid and could not
            be parsed

        Returns:
            New Vocabulary with the given sources added to it
        """

        # create a new vocabulary by reparsing the existing sources
        new_vocabulary = Vocabulary(settings=copy.copy(vocabulary.settings))
        parser = RdfParser()
        for source in vocabulary.sources.values():
            source_copy = copy.deepcopy(source)
            source_copy.clear()
            parser.parse_source_into_vocabulary(
                source=source_copy, vocabulary=new_vocabulary
            )

        # try to parse in the new sources and post_process
        try:
            for source in sources:
                parser.parse_source_into_vocabulary(
                    source=source, vocabulary=new_vocabulary
                )
            PostProcessor.post_process_vocabulary(
                vocabulary=new_vocabulary, old_vocabulary=vocabulary
            )
        except Exception as e:
            raise ParsingException(e.args)

        return new_vocabulary

    @classmethod
    def is_label_blacklisted(cls, label: str) -> bool:
        """Checks if the given label is forbidden for an entity to possess

        Args:
            label (str): label to check

        Returns:
            bool
        """
        return label in label_blacklist

    @classmethod
    def is_label_illegal(cls, label: str) -> bool:
        """Checks if the given label contains a forbidden char

        Args:
            label (str): label to check

        Returns:
            bool, True if label forbidden
        """
        for c in label:
            if c not in label_char_whitelist:
                return True
        return False

    @classmethod
    def get_label_conflicts_in_vocabulary(cls, vocabulary: Vocabulary) -> LabelSummary:
        """
        Compute a summary for all labels present in the vocabulary.
        The summary contains all naming clashes and illegal labels.

        Args:
            vocabulary (Vocabulary): Vocabulary to analyse

        Returns:
            LabelSummary
        """

        def get_conflicts_in_group(entities_to_check: List[Dict]):
            # maps label to list of entities with that label
            used_labels: Dict[str, List[Entity]] = {}
            duplicate_labels = set()

            # process entities to find conflicts
            for entity_list in entities_to_check:
                for entity in entity_list.values():
                    label = entity.get_label()
                    if label in used_labels:
                        duplicate_labels.add(label)
                        used_labels[label].append(entity)
                    else:
                        used_labels[label] = [entity]

            # sort duplicate_labels to have alphabetical order in list
            dup_list = list(duplicate_labels)
            dup_list = sorted(dup_list, key=str.casefold)

            result: Dict[str, List[Entity]] = {}
            # store and log findings
            for label in dup_list:
                result[label] = used_labels[label]

            return result

        def get_blacklisted_labels(entities_to_check: List[Dict]):
            result: List[Tuple[str, Entity]] = []
            for entity_list in entities_to_check:
                for entity in entity_list.values():
                    label = entity.get_label()
                    if cls.is_label_blacklisted(label):
                        result.append((label, entity))

            return result

        def get_illegal_labels(entities_to_check: List[Dict]):
            result: List[Tuple[str, Entity]] = []
            for entity_list in entities_to_check:
                for entity in entity_list.values():
                    label = entity.get_label()
                    if cls.is_label_illegal(label):
                        result.append((label, entity))

            return result

        summary = LabelSummary(
            class_label_duplicates=get_conflicts_in_group(
                [
                    vocabulary.classes,
                    vocabulary.individuals,
                    vocabulary.get_enum_dataytypes(),
                ]
            ),
            field_label_duplicates=get_conflicts_in_group(
                [vocabulary.data_properties, vocabulary.object_properties]
            ),
            datatype_label_duplicates=get_conflicts_in_group([vocabulary.datatypes]),
            blacklisted_labels=get_blacklisted_labels(
                [
                    vocabulary.classes,
                    vocabulary.individuals,
                    vocabulary.data_properties,
                    vocabulary.object_properties,
                ]
            ),
            labels_with_illegal_chars=get_illegal_labels(
                [
                    vocabulary.classes,
                    vocabulary.individuals,
                    vocabulary.data_properties,
                    vocabulary.object_properties,
                    vocabulary.datatypes,
                ]
            ),
        )

        return summary

    @classmethod
    def is_vocabulary_valid(cls, vocabulary: Vocabulary) -> bool:
        """
        Test if the given vocabulary is valid -> all labels are unique and
        correct

        Args:
            vocabulary (Vocabulary): Vocabulary to analyse

        Returns:
            bool
        """
        return VocabularyConfigurator.get_label_conflicts_in_vocabulary(
            vocabulary
        ).is_valid()

    @classmethod
    def get_missing_dependency_statements(
        cls, vocabulary: Vocabulary
    ) -> List[DependencyStatement]:
        """
        Get a list of all Dependencies that are currently missing in the
        vocabulary in form of DependencyStatements

        Args:
            vocabulary (Vocabulary): Vocabulary to analyse

        Returns:
            List[DependencyStatement]
        """
        missing_dependencies: List[DependencyStatement] = []
        for source in vocabulary.get_source_list():
            for statement in source.dependency_statements:
                if not statement.fulfilled:
                    missing_dependencies.append(statement)
        return missing_dependencies

    @classmethod
    def get_missing_dependencies(cls, vocabulary: Vocabulary) -> List[str]:
        """
        Get a list of all Dependencies that are currently missing in the
        vocabulary in form of iris

        Args:
            vocabulary (Vocabulary): Vocabulary to analyse

        Returns:
            List[str]: List of missing iris
        """

        missing_dependencies: Set[str] = set()
        for source in vocabulary.get_source_list():
            for statement in source.dependency_statements:
                if not statement.fulfilled:
                    missing_dependencies.add(statement.dependency_iri)
        return list(missing_dependencies)

    @classmethod
    def get_parsing_logs(cls, vocabulary: Vocabulary) -> List[ParsingError]:
        """
        Get the parsing logs of a vocabulary

        Args:
            vocabulary (Vocabulary): Vocabulary to analyse

        Returns:
            List[ParsingError]
        """
        res = []
        for source in vocabulary.sources.values():
            res.extend(source.get_parsing_log(vocabulary))
        return res

    @classmethod
    def generate_vocabulary_models(
        cls,
        vocabulary: Vocabulary,
        path: Optional[str] = None,
        filename: Optional[str] = None,
        alternative_manager_name: Optional[str] = None,
    ) -> Optional[str]:
        """
        Export the given vocabulary as python model file.
        All vocabulary classes will be converted to python classes,
        with their CRs as property fields.
        If path and filename are given, the generated file will be saved,
        else the file content is returned as string.

        Args:
            vocabulary (Vocabulary): Vocabulary to export
            path (Optional[str]): Path where the file should be saved
            filename (Optional[str]): Name of the file
            alternative_manager_name (Optional[str]): alternative name for
                the semantic_manager. The manager of the model can than also
                be referenced over the object with this name

        Raises:
            Exception: if file can not be saved as specified with path and
                       filename
            Exception: if vocabulary has label conflicts and is thus not valid

        Returns:
            Optional[str], generated content if path or filename not given
        """

        if not cls.is_vocabulary_valid(vocabulary):
            raise Exception(
                "Vocabulary was not valid. Label conflicts "
                "prevented the generation of models. Check for conflicts with: "
                "VocabularyConfigurator."
                "get_label_conflicts_in_vocabulary(vocabulary)"
            )

        def split_string_into_lines(string: str, limit: int) -> [str]:
            """Helper methode, takes a long string and splits it into
            multiple parts that each represent one line

            Args:
                string: value to split
                limit: line limit
            Returns:
                [str], string separated into lines
            """
            last_space_index = 0
            last_split_index = 0
            current_index = 0
            result = []

            for char in string:
                if char == " ":
                    last_space_index = current_index
                if current_index - last_split_index > limit:
                    result.append(string[last_split_index:last_space_index])
                    last_split_index = last_space_index + 1
                current_index += 1

            # add the remaining part, if the last character of the string was
            # not a space at the perfect position
            if not last_split_index == len(string):
                result.append(string[last_split_index:current_index])
            return result

        content: str = (
            '"""\nAutogenerated Models for the vocabulary '
            "described "
            "by the ontologies:\n"
        )
        for source in vocabulary.sources.values():
            if not source.predefined:
                content += f"\t{source.ontology_iri} ({source.source_name})\n"
        content += '"""\n\n'

        # imports
        content += "from enum import Enum\n"
        content += "from typing import Dict, Union, List\n"
        content += (
            "from filip.semantics.semantics_models import\\"
            "\n\tSemanticClass,\\"
            "\n\tSemanticIndividual,\\"
            "\n\tRelationField,\\"
            "\n\tDataField,\\"
            "\n\tSemanticDeviceClass,\\"
            "\n\tDeviceAttributeField,\\"
            "\n\tCommandField"
        )
        content += "\n"
        content += (
            "from filip.semantics.semantics_manager import\\"
            "\n\tSemanticsManager,\\"
            "\n\tInstanceRegistry"
        )

        content += "\n\n\n"
        content += f"semantic_manager: SemanticsManager = SemanticsManager("
        content += "\n\t"
        content += "instance_registry=InstanceRegistry(),"
        content += "\n"
        content += ")"
        content += "\n\n"
        if alternative_manager_name is not None:
            content += f"{alternative_manager_name}: SemanticsManager"
            content += f"= semantic_manager"
            content += "\n\n"
        content += "# ---------CLASSES--------- #"

        # the classes need to be added in order, so that the parents are
        # defined, the moment the children are added
        classes: List[Class] = vocabulary.get_classes_sorted_by_label()
        class_order: List[Class] = []
        children: Dict[str, Set] = {}
        added_class_iris = set()

        # set up data for computation of order
        iri_queue = ["http://www.w3.org/2002/07/owl#Thing"]
        for class_ in classes:
            if class_.iri not in children:
                children[class_.iri] = set()

                if class_.label == "Currency":
                    print(class_.get_parent_classes(vocabulary))

            for parent in class_.get_parent_classes(vocabulary):
                if parent.iri not in children:
                    children[parent.iri] = set()
                children[parent.iri].add(class_.iri)

        # compute class order, in the queue are always the classes, that have
        # all parents already defined (starting with Thing).
        # It is added from the queue and all children who are now fully
        # defined are added to the queue
        while len(iri_queue) > 0:
            # remove from queue
            parent_iri = iri_queue[0]
            del iri_queue[0]

            # add to class_order
            parent = vocabulary.classes[parent_iri]
            class_order.append(parent)
            added_class_iris.add(parent_iri)

            # check children
            child_iris = children[parent_iri]
            for child_iri in child_iris:
                child = vocabulary.classes[child_iri]

                # all parents added, add child to queue
                if len(
                    [p for p in child.parent_class_iris if p in added_class_iris]
                ) == len(child.parent_class_iris):

                    if not child_iri in added_class_iris:
                        iri_queue.append(child_iri)

        for class_ in class_order:

            content += "\n\n\n"
            # Parent Classes
            parent_class_string = ""
            parents = class_.get_parent_classes(vocabulary, remove_redundancy=True)

            # Device Class, only add if this is a device class and it was not
            # added for a parent
            if class_.is_iot_class(vocabulary):
                if True not in [p.is_iot_class(vocabulary) for p in parents]:
                    parent_class_string = " ,SemanticDeviceClass"

            for parent in parents:
                parent_class_string += f", {parent.get_label()}"

            parent_class_string = parent_class_string[
                2:
            ]  # remove first comma and space
            if parent_class_string == "":
                parent_class_string = "SemanticClass"

            content += f"class {class_.get_label()}({parent_class_string}):"

            # add class docstring
            content += f'\n\t"""'
            for line in split_string_into_lines(class_.comment, 75):
                content += f"\n\t{line}"
            if class_.comment == "":
                content += "\n\tGenerated SemanticClass without description"
            content += f"\n\n\t"
            content += f"Source(s): \n\t\t"

            for source_id in class_.source_ids:
                content += (
                    f"{vocabulary.sources[source_id].ontology_iri} "
                    f"({vocabulary.sources[source_id].source_name})"
                )
            content += f'\n\t"""'

            # ------Constructors------
            if class_.get_label() == "Thing":
                content += "\n\n\t"
                content += "def __new__(cls, *args, **kwargs):"
                content += "\n\t\t"
                content += f"kwargs['semantic_manager'] = semantic_manager"
                content += "\n\t\t"
                content += "return super().__new__(cls, *args, **kwargs)"

                content += "\n\n\t"
                content += "def __init__(self, *args, **kwargs):"
                content += "\n\t\t"
                content += f"kwargs['semantic_manager'] = semantic_manager"
                content += "\n\t\t"
                content += "is_initialised = 'id' in self.__dict__"
                content += "\n\t\t"
                content += "super().__init__(*args, **kwargs)"

            else:
                content += "\n\n\t"
                content += "def __init__(self, *args, **kwargs):"
                content += "\n\t\t"
                content += "is_initialised = 'id' in self.__dict__"
                content += "\n\t\t"
                content += "super().__init__(*args, **kwargs)"

            # ------Init Fields------
            content += "\n\t\t"
            content += "if not is_initialised:"
            # Only initialise values once
            for cdr in class_.get_combined_data_relations(vocabulary):
                if not cdr.is_device_relation(vocabulary):
                    content += "\n\t\t\t"
                    content += (
                        f"self."
                        f"{cdr.get_property_label(vocabulary)}._rules = "
                        f"{cdr.export_rule(vocabulary, stringify_fields=True)}"
                    )

            if len(class_.get_combined_object_relations(vocabulary)) > 0:
                content += "\n"
            for cor in class_.get_combined_object_relations(vocabulary):
                content += "\n\t\t\t"
                content += (
                    f"self."
                    f"{cor.get_property_label(vocabulary)}._rules = "
                    f"{cor.export_rule(vocabulary, stringify_fields=False)}"
                )

            if len(class_.get_combined_relations(vocabulary)) > 0:
                content += "\n"
            for cr in class_.get_combined_relations(vocabulary):
                content += "\n\t\t\t"
                content += (
                    f"self.{cr.get_property_label(vocabulary)}"
                    f"._instance_identifier = "
                    f"self.get_identifier()"
                )

            # ------Add preset Values------
            for cdr in class_.get_combined_data_relations(vocabulary):
                # Add fixed values to fields, for CDRs these values need to be
                # strings. Only add the statement on the uppermost occurring
                # class
                if not cdr.is_device_relation(vocabulary):
                    for rel in cdr.get_relations(vocabulary):
                        if rel.id in class_.relation_ids:
                            # only reinitialise the fields if this child class
                            # changed them
                            if rel.restriction_type == RestrictionType.value:
                                content += "\n\t\t\t"
                                content += (
                                    f"self."
                                    f"{cdr.get_property_label(vocabulary)}"
                                    f".add("
                                    f"'{rel.target_statement.target_data_value}')"
                                )

            if len(class_.get_combined_object_relations(vocabulary)) > 0:
                content += "\n"
            for cor in class_.get_combined_object_relations(vocabulary):
                # Add fixed values to fields, for CORs these values need to be
                # Individuals.
                # Only add the statement on the uppermost occurring class
                for rel in cor.get_relations(vocabulary):
                    if rel.id in class_.relation_ids:
                        i = vocabulary.get_label_for_entity_iri(rel.get_targets()[0][0])
                        if rel.restriction_type == RestrictionType.value:
                            content += "\n\t\t\t"
                            content += (
                                f"self."
                                f"{cor.get_property_label(vocabulary)}"
                                f".add({i}())"
                            )

            # if no content was added af the not initialised if, removed it
            # again, and its preceding \n
            if content[-22:] == "if not is_initialised:":
                content = content[:-25]

            # make space the same for each case above
            if "\n" in content[-2:]:
                content = content[:-1]

            def build_field_comment(cr: CombinedRelation) -> str:
                comment = vocabulary.get_entity_by_iri(cr.property_iri).comment
                res = ""
                if comment != "":
                    res += f'\n\t"""'
                    for line in split_string_into_lines(comment, 75):
                        res += f"\n\t{line}"
                    res += f'\n\t"""'
                return res

            # ------Add Data Fields------
            if len(class_.get_combined_data_relations(vocabulary)) > 0:
                content += "\n\n\t"
                content += "# Data fields"
            for cdr in class_.get_combined_data_relations(vocabulary):
                cdr_type = cdr.get_field_type(vocabulary)
                if cdr_type == DataFieldType.simple:
                    content += "\n\n\t"
                    label = cdr.get_property_label(vocabulary)
                    content += f"{label}: DataField = DataField("
                    content += "\n\t\t"
                    content += f"name='{label}',"
                    content += "\n\t\t"
                    content += (
                        f"rule='"
                        f"{cdr.get_all_targetstatements_as_string(vocabulary)}',"
                    )
                    content += "\n\t\t"
                    content += f"semantic_manager=semantic_manager)"
                    content += build_field_comment(cdr)

                elif cdr_type == DataFieldType.command:
                    content += "\n\n\t"
                    label = cdr.get_property_label(vocabulary)
                    content += f"{label}: CommandField = CommandField("
                    content += "\n\t\t"
                    content += f"name='{label}',"
                    content += "\n\t\t"
                    content += f"semantic_manager=semantic_manager)"
                    content += build_field_comment(cdr)

                elif cdr_type == DataFieldType.device_attribute:
                    content += "\n\n\t"
                    label = cdr.get_property_label(vocabulary)
                    content += (
                        f"{label}: DeviceAttributeField " f"= DeviceAttributeField("
                    )
                    content += "\n\t\t"
                    content += f"name='{label}',"
                    content += "\n\t\t"
                    content += f"semantic_manager=semantic_manager)"
                    content += build_field_comment(cdr)

            # ------Add Relation Fields------
            if len(class_.get_combined_object_relations(vocabulary)) > 0:
                content += "\n\n\t"
                content += "# Relation fields"
            for cor in class_.get_combined_object_relations(vocabulary):
                content += "\n\n\t"
                label = cor.get_property_label(vocabulary)
                content += f"{label}: RelationField = RelationField("
                content += "\n\t\t"
                content += f"name='{label}',"
                content += "\n\t\t"
                content += (
                    f"rule='" f"{cor.get_all_targetstatements_as_string(vocabulary)}',"
                )
                content += "\n\t\t"
                if not len(cor.get_inverse_of_labels(vocabulary)) == 0:
                    content += "inverse_of="
                    content += str(cor.get_inverse_of_labels(vocabulary))
                    content += ",\n\t\t"

                content += f"semantic_manager=semantic_manager)"
                content += build_field_comment(cor)

        content += "\n\n\n"
        content += "# ---------Individuals--------- #"

        for individual in vocabulary.individuals.values():
            content += "\n\n\n"

            parent_class_string = ""
            for parent in individual.get_parent_classes(vocabulary):
                parent_class_string += f", {parent.get_label()}"
            parent_class_string = parent_class_string[2:]

            content += f"class {individual.get_label()}(SemanticIndividual):"
            content += "\n\t"
            content += f"_parent_classes: List[type] = [{parent_class_string}]"

        content += "\n\n\n"

        content += "# ---------Datatypes--------- #"
        content += "\n"

        # Datatypes catalogue
        content += f"semantic_manager.datatype_catalogue = "
        content += "{"
        for name, datatype in vocabulary.datatypes.items():
            definition = datatype.export()
            content += "\n\t"
            # content += f"'{datatype.get_label()}': \t {definition},"
            content += f"'{datatype.get_label()}': "
            content += "{\n"
            for key, value in definition.items():
                string_value = f"'{value}'" if type(value) == str else value
                content += f"\t\t'{key}': {string_value},\n"
            content += "\t},"

        content += "\n"
        content += "}"

        # Build datatypes with enums as Enums
        content += "\n\n\n"
        for datatype in vocabulary.get_enum_dataytypes().values():
            content += f"class {datatype.get_label()}(str, Enum):"
            for value in datatype.enum_values:
                content += f"\n\tvalue_{value} = '{value}'"
            content += "\n\n\n"

        content += "# ---------Class Dict--------- #"

        # build class dict
        content += "\n\n"
        content += f"semantic_manager.class_catalogue = "
        content += "{"
        for class_ in vocabulary.get_classes_sorted_by_label():
            content += "\n\t"
            content += f"'{class_.get_label()}': {class_.get_label()},"
        content += "\n\t}"
        content += "\n"

        # build individual dict
        content += "\n\n"
        content += f"semantic_manager.individual_catalogue = "
        content += "{"
        for individual in vocabulary.individuals.values():
            content += "\n\t"
            content += f"'{individual.get_label()}': {individual.get_label()},"
        content += "\n\t}"
        content += "\n"

        if path is None or filename is None:
            return content
        else:
            path = pathlib.Path(path).joinpath(filename).with_suffix(".py")

            with open(path, "w", encoding="utf-8") as text_file:
                text_file.write(content)


class ParsingException(Exception):
    """Error Class that is raised if parsing of an ontology was unsuccessful"""

    # Constructor or Initializer
    def __init__(self, value):
        self.value = value

    # __str__ is to print() the value
    def __str__(self):
        return repr(self.value)



"""
NGSI LD models for context broker interaction
"""

import logging
from typing import Any, List, Dict, Union, Optional
import re
from geojson_pydantic import (
    Point,
    MultiPoint,
    LineString,
    MultiLineString,
    Polygon,
    MultiPolygon,
)
from typing_extensions import Self
from aenum import Enum
from pydantic import field_validator, ConfigDict, BaseModel, Field, model_validator
from filip.models.ngsi_v2 import ContextEntity
from filip.utils.validators import (
    FiwareRegex,
    validate_fiware_datatype_string_protect,
    validate_fiware_standard_regex,
)
from pydantic_core import ValidationError


class DataTypeLD(str, Enum):
    """
    In NGSI-LD the data types on context entities are only divided into properties and relationships.
    """

    _init_ = "value __doc__"
    GEOPROPERTY = "GeoProperty", "A property that represents a geometry value"
    PROPERTY = "Property", "All attributes that do not represent a relationship"
    RELATIONSHIP = (
        "Relationship",
        "Reference to another context entity, which can be identified with a URN.",
    )


# NGSI-LD entity models
class ContextProperty(BaseModel):
    """
    The model for a property is represented by a JSON object with the following syntax:

    The attribute value is specified by the value, whose value can be any data type. This does not need to be
    specified further.

    The NGSI type of the attribute is fixed and does not need to be specified.
    Example:

        >>> data = {"value": <...>}

        >>> attr = ContextProperty(**data)

    """

    model_config = ConfigDict(extra="allow")  # In order to allow nested properties
    type: Optional[str] = Field(default="Property", title="type", frozen=True)
    value: Optional[
        Union[
            Union[float, int, bool, str, List, Dict[str, Any]],
            List[Union[float, int, bool, str, List, Dict[str, Any]]],
        ]
    ] = Field(default=None, title="Property value", description="the actual data")
    observedAt: Optional[str] = Field(
        None,
        title="Timestamp",
        description="Representing a timestamp for the "
        "incoming value of the property.",
        max_length=256,
        min_length=1,
    )
    field_validator("observedAt")(validate_fiware_datatype_string_protect)

    createdAt: Optional[str] = Field(
        None,
        title="Timestamp",
        description="Representing a timestamp for the "
        "creation time of the property.",
        max_length=256,
        min_length=1,
    )
    field_validator("createdAt")(validate_fiware_datatype_string_protect)

    modifiedAt: Optional[str] = Field(
        None,
        title="Timestamp",
        description="Representing a timestamp for the "
        "last modification of the property.",
        max_length=256,
        min_length=1,
    )
    field_validator("modifiedAt")(validate_fiware_datatype_string_protect)

    unitCode: Optional[str] = Field(
        None,
        title="Unit Code",
        description="Representing the unit of the value. "
        "Should be part of the defined units "
        "by the UN/ECE Recommendation No. 21"
        "https://unece.org/fileadmin/DAM/cefact/recommendations/rec20/rec20_rev3_Annex2e.pdf ",
        max_length=256,
        min_length=1,
    )
    field_validator("unitCode")(validate_fiware_datatype_string_protect)

    datasetId: Optional[str] = Field(
        None,
        title="dataset Id",
        description="It allows identifying a set or group of property values",
        max_length=256,
        min_length=1,
    )
    field_validator("datasetId")(validate_fiware_datatype_string_protect)

    @classmethod
    def get_model_fields_set(cls):
        """
        Get all names and aliases of the model fields.
        """
        return set(
            [field.validation_alias for (_, field) in cls.model_fields.items()]
            + [field_name for field_name in cls.model_fields]
        )

    @field_validator("type")
    @classmethod
    def check_property_type(cls, value):
        """
        Force property type to be "Property"
        Args:
            value: value field
        Returns:
            value
        """
        valid_property_types = ["Property", "Relationship", "TemporalProperty"]
        if value not in valid_property_types:
            msg = (
                f"NGSI_LD Properties must have type {valid_property_types}, "
                f'not "{value}"'
            )
            logging.warning(msg=msg)
            raise ValueError(msg)
        return value


class NamedContextProperty(ContextProperty):
    """
    Context properties are properties of context entities. For example, the current speed of a car could be modeled
    as the current_speed property of the car-104 entity.

    In the NGSI-LD data model, properties have a name, the type "property" and a value.
    """

    name: str = Field(
        title="Property name",
        description="The property name describes what kind of property the "
        "attribute value represents of the entity, for example "
        "current_speed. Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
    )
    field_validator("name")(validate_fiware_datatype_string_protect)


class ContextGeoPropertyValue(BaseModel):
    """
    The value for a Geo property is represented by a JSON object with the following syntax:

    A type with value "Point" and the
    coordinates with a list containing the coordinates as value

    Example:
        "value": {
            "type": "Point",
            "coordinates": [
                -3.80356167695194,
                43.46296641666926
            ]
        }
    }

    """

    type: Optional[str] = Field(default=None, title="type", frozen=True)
    model_config = ConfigDict(extra="allow")

    @model_validator(mode="after")
    def check_geoproperty_value(self) -> Self:
        """
        Check if the value is a valid GeoProperty
        """
        if self.model_dump().get("type") == "Point":
            return Point(**self.model_dump())
        elif self.model_dump().get("type") == "LineString":
            return LineString(**self.model_dump())
        elif self.model_dump().get("type") == "Polygon":
            return Polygon(**self.model_dump())
        elif self.model_dump().get("type") == "MultiPoint":
            return MultiPoint(**self.model_dump())
        elif self.model_dump().get("type") == "MultiLineString":
            return MultiLineString(**self.model_dump())
        elif self.model_dump().get("type") == "MultiPolygon":
            return MultiPolygon(**self.model_dump())
        elif self.model_dump().get("type") == "GeometryCollection":
            raise ValueError("GeometryCollection is not supported")


class ContextGeoProperty(BaseModel):
    """
    The model for a Geo property is represented by a JSON object with the following syntax:

    The attribute value is a JSON object with two contents.

    Example:

        {
        "type": "GeoProperty",
        "value": {
            "type": "Point",
            "coordinates": [
                -3.80356167695194,
                43.46296641666926
            ]
        }

    """

    model_config = ConfigDict(extra="allow")
    type: Optional[str] = Field(default="GeoProperty", title="type", frozen=True)
    value: Optional[
        Union[
            ContextGeoPropertyValue,
            Point,
            LineString,
            Polygon,
            MultiPoint,
            MultiPolygon,
            MultiLineString,
        ]
    ] = Field(default=None, title="GeoProperty value", description="the actual data")
    observedAt: Optional[str] = Field(
        default=None,
        title="Timestamp",
        description="Representing a timestamp for the "
        "incoming value of the property.",
        max_length=256,
        min_length=1,
    )
    field_validator("observedAt")(validate_fiware_datatype_string_protect)

    datasetId: Optional[str] = Field(
        None,
        title="dataset Id",
        description="It allows identifying a set or group of property values",
        max_length=256,
        min_length=1,
    )
    field_validator("datasetId")(validate_fiware_datatype_string_protect)


class NamedContextGeoProperty(ContextGeoProperty):
    """
    Context GeoProperties are geo properties of context entities. For example, the coordinates of a building .

    In the NGSI-LD data model, properties have a name, the type "Geoproperty" and a value.
    """

    name: str = Field(
        title="Property name",
        description="The property name describes what kind of property the "
        "attribute value represents of the entity, for example "
        "current_speed. Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
    )
    field_validator("name")(validate_fiware_datatype_string_protect)


class ContextRelationship(BaseModel):
    """
    The model for a relationship is represented by a JSON object with the following syntax:

    The attribute value is specified by the object, whose value can be a reference to another context entity. This
    should be specified as the URN. The existence of this entity is not assumed.

    The NGSI type of the attribute is fixed and does not need to be specified.

    Example:

        >>> data = {"object": <...>}

        >>> attr = ContextRelationship(**data)

    """

    model_config = ConfigDict(extra="allow")  # In order to allow nested relationships
    type: Optional[str] = Field(default="Relationship", title="type", frozen=True)
    object: Optional[
        Union[
            Union[float, int, bool, str, List, Dict[str, Any]],
            List[Union[float, int, bool, str, List, Dict[str, Any]]],
        ]
    ] = Field(
        default=None, title="Realtionship object", description="the actual object id"
    )

    datasetId: Optional[str] = Field(
        None,
        title="dataset Id",
        description="It allows identifying a set or group of property values",
        max_length=256,
        min_length=1,
    )
    field_validator("datasetId")(validate_fiware_datatype_string_protect)

    observedAt: Optional[str] = Field(
        None,
        titel="Timestamp",
        description="Representing a timestamp for the "
        "incoming value of the property.",
        max_length=256,
        min_length=1,
    )
    field_validator("observedAt")(validate_fiware_datatype_string_protect)

    @field_validator("type")
    @classmethod
    def check_relationship_type(cls, value):
        """
        Force property type to be "Relationship"
        Args:
            value: value field
        Returns:
            value
        """
        if not value == "Relationship":
            logging.warning(msg='NGSI_LD relationships must have type "Relationship"')
        value = "Relationship"
        return value


class NamedContextRelationship(ContextRelationship):
    """
    Context Relationship are relations of context entities to each other.
    For example, the current_speed of the entity car-104 could be modeled.
    The location could be modeled as located_in the entity Room-001.

    In the NGSI-LD data model, relationships have a name, the type "relationship" and an object.
    """

    name: str = Field(
        title="Attribute name",
        description="The attribute name describes what kind of property the "
        "attribute value represents of the entity, for example "
        "current_speed. Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
        # pattern=FiwareRegex.string_protect.value,
        # Make it FIWARE-Safe
    )
    field_validator("name")(validate_fiware_datatype_string_protect)


class ContextLDEntityBase(BaseModel):
    """
    Base Model for an entity is represented by a JSON object with the following
    syntax.

    The entity id is specified by the object's id property, whose value
    is a string containing the entity id.

    The entity type is specified by the object's type property, whose value
    is a string containing the entity's type name.
    """

    model_config = ConfigDict(
        extra="allow", validate_default=True, validate_assignment=True
    )
    id: str = Field(
        ...,
        title="Entity Id",
        description="Id of an entity in an NGSI context broker. Allowed "
        "characters are the ones in the plain ASCII set, except "
        "the following ones: control characters, "
        "whitespace, &, ?, / and #."
        "the id should be structured according to the urn naming scheme.",
        json_schema_extra={"example": "urn:ngsi-ld:Room:001"},
        max_length=256,
        min_length=1,
        # pattern=FiwareRegex.standard.value,  # Make it FIWARE-Safe
        frozen=True,
    )
    field_validator("id")(validate_fiware_standard_regex)
    type: str = Field(
        ...,
        title="Entity Type",
        description="Id of an entity in an NGSI context broker. "
        "Allowed characters are the ones in the plain ASCII set, "
        "except the following ones: control characters, "
        "whitespace, &, ?, / and #.",
        json_schema_extra={"example": "Room"},
        max_length=256,
        min_length=1,
        # pattern=FiwareRegex.standard.value,  # Make it FIWARE-Safe
        frozen=True,
    )
    field_validator("type")(validate_fiware_standard_regex)
    context: Optional[Union[str, List[str], Dict]] = Field(
        title="@context",
        default=None,
        description="The @context in JSON-LD is used to expand terms, provided as short "
        "hand strings, to concepts, specified as URIs, and vice versa, "
        "to compact URIs into terms "
        "The main implication of NGSI-LD API is that if the @context is "
        "a compound one, i.e. an @context which references multiple "
        "individual @context, served by resources behind different URIs, "
        "then a wrapper @context has to be created and hosted.",
        examples=["https://n5geh.github.io/n5geh.test-context.io/context_saref.jsonld"],
        alias="@context",
        validation_alias="@context",
        frozen=False,
    )


class ContextLDEntityKeyValues(ContextLDEntityBase):
    """
    Base Model for an entity is represented by a JSON object with the following
    syntax.

    The entity id is specified by the object's id property, whose value
    is a string containing the entity id.

    The entity type is specified by the object's type property, whose value
    is a string containing the entity's type name.

    """

    model_config = ConfigDict(
        extra="allow", validate_default=True, validate_assignment=True
    )

    def to_entity(self):
        """
        Convert the entity to a normalized representation.
        """
        return ContextLDEntity(
            **{
                "id": self.id,
                "type": self.type,
                "context": self.context if self.context else None,
                **{
                    key: {
                        "type": "Property",
                        "value": value,
                    }
                    for key, value in self.model_dump().items()
                    if key not in ["id", "type", "context"]
                },
            }
        )


class PropertyFormat(str, Enum):
    """
    Format to decide if properties of ContextEntity class are returned as
    List of NamedContextAttributes or as Dict of ContextAttributes.
    """

    LIST = "list"
    DICT = "dict"


class ContextLDEntity(ContextLDEntityBase):
    """
    Context LD entities, or simply entities, are the center of gravity in the
    FIWARE NGSI-LD information model. An entity represents a thing, i.e., any
    physical or logical object (e.g., a sensor, a person, a room, an issue in
    a ticketing system, etc.). Each entity has an entity id.
    Furthermore, the type system of FIWARE NGSI enables entities to have an
    entity type. Entity types are semantic types; they are intended to describe
    the type of thing represented by the entity. For example, a context
    entity #with id sensor-365 could have the type temperatureSensor.

    Each entity is uniquely identified by its id.

    The entity id is specified by the object's id property, whose value
    is a string containing the entity id.

    The entity type is specified by the object's type property, whose value
    is a string containing the entity's type name.

    Entity attributes are specified by additional properties and relationships, whose names are
    the name of the attribute and whose representation is described in the
    "ContextProperty"/"ContextRelationship"-model. Obviously, id and type are
    not allowed to be used as attribute names.

    Example:

        >>> data = {'id': 'MyId',
                    'type': 'MyType',
                    'my_attr': {'value': 20}}

        >>> entity = ContextLDEntity(**data)

    """

    model_config = ConfigDict(
        extra="allow",
        validate_default=True,
        validate_assignment=True,
        populate_by_name=True,
    )

    observationSpace: Optional[ContextGeoProperty] = Field(
        default=None,
        title="Observation Space",
        description="The geospatial Property representing "
        "the geographic location that is being "
        "observed, e.g. by a sensor. "
        "For example, in the case of a camera, "
        "the location of the camera and the "
        "observationspace are different and "
        "can be disjoint. ",
    )

    @field_validator("context")
    @classmethod
    def return_context(cls, context):
        return context

    operationSpace: Optional[ContextGeoProperty] = Field(
        default=None,
        title="Operation Space",
        description="The geospatial Property representing "
        "the geographic location in which an "
        "Entity,e.g. an actuator is active. "
        "For example, a crane can have a "
        "certain operation space.",
    )

    createdAt: Optional[str] = Field(
        None,
        title="Timestamp",
        description="Representing a timestamp for the "
        "creation time of the property.",
        max_length=256,
        min_length=1,
    )
    field_validator("createdAt")(validate_fiware_datatype_string_protect)

    modifiedAt: Optional[str] = Field(
        None,
        title="Timestamp",
        description="Representing a timestamp for the "
        "last modification of the property.",
        max_length=256,
        min_length=1,
    )
    field_validator("modifiedAt")(validate_fiware_datatype_string_protect)

    def __init__(self, **data):
        # There is currently no validation for extra fields
        data.update(self._validate_attributes(data))
        super().__init__(**data)

    @classmethod
    def get_model_fields_set(cls):
        """
        Get all names and aliases of the model fields.
        """
        return set(
            [field.validation_alias for (_, field) in cls.model_fields.items()]
            + [field_name for field_name in cls.model_fields]
        )

    @classmethod
    def _validate_single_property(
        cls, attr
    ) -> Union[ContextProperty, ContextRelationship, ContextGeoProperty]:
        # skip validation if pre-defined model is already used
        if type(attr) in [ContextProperty, ContextRelationship, ContextGeoProperty]:
            return attr
        property_fields = ContextProperty.get_model_fields_set()
        property_fields.remove(None)
        # subattrs = {}
        if attr.get("type") == "Relationship":
            attr_instance = ContextRelationship.model_validate(attr)
        elif attr.get("type") == "GeoProperty":
            try:
                attr_instance = ContextGeoProperty.model_validate(attr)
            except Exception as e:
                pass
        elif attr.get("type") == "Property" or attr.get("type") is None:
            attr_instance = ContextProperty.model_validate(attr)
        else:
            raise ValueError(f"Attribute {attr.get('type')} " "is not a valid type")
        for subkey, subattr in attr.items():
            if isinstance(subattr, dict) and subkey not in property_fields:
                attr_instance.model_extra.update(
                    {subkey: cls._validate_single_property(attr=subattr)}
                )
        return attr_instance

    @classmethod
    def _validate_attributes(cls, data: Dict):
        entity_fields = cls.get_model_fields_set()
        entity_fields.remove(None)
        # Initialize the attribute dictionary
        attrs = {}
        # Iterate through the data
        for key, attr in data.items():
            # Check if the keyword is not already present in the fields
            if key not in entity_fields:
                attrs[key] = cls._validate_single_property(attr=attr)
        return attrs

    def model_dump(self, *args, by_alias: bool = True, **kwargs):
        return super().model_dump(*args, by_alias=by_alias, **kwargs)

    @field_validator("id")
    @classmethod
    def _validate_id(cls, id: str):
        if not id.startswith("urn:ngsi-ld:"):
            logging.warning(
                msg="It is recommended that the entity id to be a URN,"
                'starting with the namespace "urn:ngsi-ld:"'
            )
        return id

    def get_properties(
        self, response_format: Union[str, PropertyFormat] = PropertyFormat.LIST
    ) -> Union[List[NamedContextProperty], Dict[str, ContextProperty]]:
        """
        Get all properties of the entity.
        Args:
            response_format:

        Returns:

        """
        response_format = PropertyFormat(response_format)
        # response format dict:
        if response_format == PropertyFormat.DICT:
            final_dict = {}
            for key, value in self.model_dump(exclude_unset=True).items():
                if key not in ContextLDEntity.get_model_fields_set():
                    if value.get("type") != DataTypeLD.RELATIONSHIP:
                        if value.get("type") == DataTypeLD.GEOPROPERTY:
                            final_dict[key] = ContextGeoProperty(**value)
                        elif value.get("type") == DataTypeLD.PROPERTY:
                            final_dict[key] = ContextProperty(**value)
                        else:  # named context property by default
                            final_dict[key] = ContextProperty(**value)
            return final_dict
        # response format list:
        final_list = []
        for key, value in self.model_dump(exclude_unset=True).items():
            if key not in ContextLDEntity.get_model_fields_set():
                if value.get("type") != DataTypeLD.RELATIONSHIP:
                    if value.get("type") == DataTypeLD.GEOPROPERTY:
                        final_list.append(NamedContextGeoProperty(name=key, **value))
                    elif value.get("type") == DataTypeLD.PROPERTY:
                        final_list.append(NamedContextProperty(name=key, **value))
                    else:  # named context property by default
                        final_list.append(NamedContextProperty(name=key, **value))
        return final_list

    def delete_relationships(self, relationships: List[str]):
        """
        Delete the given relationships from the entity

        Args:
            relationships: List of relationship names

        Returns:

        """
        all_relationships = self.get_relationships(response_format="dict")
        for relationship in relationships:
            # check they are relationships
            if relationship not in all_relationships:
                raise ValueError(f"Relationship {relationship} does not exist")
            delattr(self, relationship)

    def delete_properties(
        self,
        props: Union[Dict[str, ContextProperty], List[NamedContextProperty], List[str]],
    ):
        """
        Delete the given properties from the entity

        Args:
            props: can be given in multiple forms
                1) Dict: {"<property_name>": ContextProperty, ...}
                2) List: [NamedContextProperty, ...]
                3) List: ["<property_name>", ...]

        Returns:

        """
        names: List[str] = []
        if isinstance(props, list):
            for entry in props:
                if isinstance(entry, str):
                    names.append(entry)
                elif isinstance(entry, NamedContextProperty):
                    names.append(entry.name)
        else:
            names.extend(list(props.keys()))

        # check there are no relationships
        relationship_names = [rel.name for rel in self.get_relationships()]
        for name in names:
            if name in relationship_names:
                raise TypeError(f"{name} is a relationship")

        for name in names:
            delattr(self, name)

    def add_geo_properties(
        self, attrs: Union[Dict[str, ContextGeoProperty], List[NamedContextGeoProperty]]
    ) -> None:
        """
        Add property to entity
        Args:
            attrs:
        Returns:
            None
        """
        if isinstance(attrs, list):
            attrs = {
                attr.name: ContextGeoProperty(
                    **attr.model_dump(exclude={"name"}, exclude_unset=True)
                )
                for attr in attrs
            }
        for key, attr in attrs.items():
            self.__setattr__(name=key, value=attr)

    def add_properties(
        self, attrs: Union[Dict[str, ContextProperty], List[NamedContextProperty]]
    ) -> None:
        """
        Add property to entity
        Args:
            attrs:
        Returns:
            None
        """
        if isinstance(attrs, list):
            attrs = {
                attr.name: ContextProperty(
                    **attr.model_dump(exclude={"name"}, exclude_unset=True)
                )
                for attr in attrs
            }
        for key, attr in attrs.items():
            self.__setattr__(name=key, value=attr)

    def add_relationships(
        self,
        relationships: Union[
            Dict[str, ContextRelationship], List[NamedContextRelationship]
        ],
    ) -> None:
        """
        Add relationship to entity
        Args:
            relationships:
        Returns:
            None
        """
        if isinstance(relationships, list):
            relationships = {
                attr.name: ContextRelationship(**attr.dict(exclude={"name"}))
                for attr in relationships
            }
        for key, attr in relationships.items():
            self.__setattr__(name=key, value=attr)

    def get_relationships(
        self, response_format: Union[str, PropertyFormat] = PropertyFormat.LIST
    ) -> Union[List[NamedContextRelationship], Dict[str, ContextRelationship]]:
        """
        Get all relationships of the context entity

        Args:
            response_format:

        Returns:

        """
        response_format = PropertyFormat(response_format)
        # response format dict:
        if response_format == PropertyFormat.DICT:
            final_dict = {}
            for key, value in self.model_dump(exclude_unset=True).items():
                if key not in ContextLDEntity.get_model_fields_set():
                    try:
                        if value.get("type") == DataTypeLD.RELATIONSHIP:
                            final_dict[key] = ContextRelationship(**value)
                    except AttributeError:  # if context attribute
                        if isinstance(value, list):
                            pass
            return final_dict
        # response format list:
        final_list = []
        for key, value in self.model_dump(exclude_unset=True).items():
            if key not in ContextLDEntity.get_model_fields_set():
                if value.get("type") == DataTypeLD.RELATIONSHIP:
                    final_list.append(NamedContextRelationship(name=key, **value))
        return final_list

    def get_context(self):
        """
        Args:
            response_format:

        Returns: context of the entity as list

        """
        _, context = self.model_dump(include={"context"}).popitem()
        if not context:
            logging.warning("No context in entity")
            return None
        else:
            return context

    def to_keyvalues(self) -> ContextLDEntityKeyValues:
        props = self.get_properties()
        rels = self.get_relationships()
        result = dict[str, Any]()
        for prop in props:
            result[prop.name] = prop.value
        for rel in rels:
            result[rel.name] = rel.object
        return ContextLDEntityKeyValues(id=self.id, type=self.type, **result)


class ActionTypeLD(str, Enum):
    """
    Options for queries
    """

    CREATE = "create"
    UPSERT = "upsert"
    UPDATE = "update"
    DELETE = "delete"


class UpdateLD(BaseModel):
    """
    Model for update action
    """

    entities: List[Union[ContextLDEntity, ContextLDEntityKeyValues]] = Field(
        description="an array of entities, each entity specified using the "
        "JSON entity representation format "
    )


class MessageLD(BaseModel):
    """
    Model for a notification message, when sent to other NGSIv2-APIs
    """

    subscriptionId: Optional[str] = Field(
        default=None,
        description="Id of the subscription the notification comes from",
    )
    data: List[ContextLDEntity] = Field(
        description="is an array with the notification data itself which "
        "includes the entity and all concerned attributes. Each "
        "element in the array corresponds to a different entity. "
        "By default, the entities are represented in normalized "
        "mode. However, using the attrsFormat modifier, a "
        "simplified representation mode can be requested."
    )



"""
Shared data models
"""

from aenum import Enum
from pydantic import ConfigDict, BaseModel, Field, field_validator, computed_field

from filip.utils.validators import validate_fiware_service_path, validate_fiware_service

core_context = "https://uri.etsi.org/ngsi-ld/v1/ngsi-ld-core-context-v1.6.jsonld"


class NgsiVersion(str, Enum):
    """
    Version of NGSI-Specification that should be used within the target system.
    Note:
        Currently, the library only supports functionality for NGSI-v2
    """

    v2 = "v2"
    ld = "ld"


class DataType(str, Enum):
    """
    When possible reuse schema.org data types
    (Text, Number, DateTime, StructuredValue, etc.).
    Remember that null is not allowed in NGSI-LD and
    therefore should be avoided as a value.

    https://schema.org/DataType
    """

    _init_ = "value __doc__"

    BOOLEAN = "Boolean", "True or False."
    DATE = "Date", "A date value in ISO 8601 date format."
    DATETIME = (
        "DateTime",
        "A combination of date and time of day in the form "
        "[-]CCYY-MM-DDThh:mm:ss[Z|(+|-)hh:mm] "
        "(see Chapter 5.4 of ISO 8601).",
    )
    NUMBER = (
        "Number",
        "Use values from 0123456789 (Unicode 'DIGIT ZERO' "
        "(U+0030) to 'DIGIT NINE' (U+0039)) rather than "
        "superficially similar Unicode symbols. Use '.' "
        "(Unicode 'FULL STOP' (U+002E)) rather than ',' to "
        "indicate a decimal point. Avoid using these symbols "
        "as a readability separator.",
    )
    INTEGER = "Integer", "Integer number"
    FLOAT = "Float", "Floating number. Please check 'DataType.Number'"
    TEXT = "Text", "https://schema.org/Text"
    TIME = (
        "Time",
        "A point in time recurring on multiple days in the form "
        "hh:mm:ss[Z|(+|-)hh:mm] (see XML schema for details).",
    )
    RELATIONSHIP = "Relationship", "Reference to another context entity"
    STRUCTUREDVALUE = "StructuredValue", (
        "Structured datatype must be an " "array or object" "serializable"
    )
    ARRAY = "Array", "Array of the types above"
    OBJECT = "Object", "JSON-Object of the types above, i.e. a dictionary"
    COMMAND = "command", "A command for IoT Devices"
    COMMAND_RESULT = (
        "commandResult",
        "An entity containing a command, "
        "contains an autogenerated attribute"
        "of this type",
    )
    COMMAND_STATUS = (
        "commandStatus",
        "An entity containing a command, "
        "contains an autogenerated attribute "
        "of this type",
    )
    GEOJSON = "geo:json", "GeoJson object"


class PaginationMethod(str, Enum):
    """
    Options for the internal pagination methods
    """

    GET = "GET"
    POST = "POST"


class FiwareHeader(BaseModel):
    """
    Define entity service paths which are supported by the NGSI
    Context Brokers to support hierarchical scopes:
    https://fiware-orion.readthedocs.io/en/master/user/service_path/index.html
    """

    model_config = ConfigDict(populate_by_name=True, validate_assignment=True)

    service: str = Field(
        alias="fiware-service",
        default="",
        max_length=50,
        description="Fiware service used for multi-tenancy",
        pattern=r"\w*$",
    )
    service_path: str = Field(
        alias="fiware-servicepath",
        default="/",
        description="Fiware service path",
        max_length=51,
    )
    valid_service = field_validator("service")(validate_fiware_service)
    valid_service_path = field_validator("service_path")(validate_fiware_service_path)


class FiwareHeaderSecure(FiwareHeader):
    """
    Defines entity service paths and an authorization via Bearer-Token which are
    supported by the NGSI
    Context Brokers to support hierarchical scopes:
    https://fiware-orion.readthedocs.io/en/master/user/service_path/index.html
    """

    authorization: str = Field(
        alias="authorization",
        default="",
        max_length=3000,
        description="authorization key",
        pattern=r".*",
    )


class LogLevel(str, Enum):
    CRITICAL = "CRITICAL"
    ERROR = "ERROR"
    WARNING = "WARNING"
    INFO = "INFO"
    DEBUG = "DEBUG"
    NOTSET = "NOTSET"

    @classmethod
    def _missing_name_(cls, name):
        """
        Class method to realize case-insensitive args

        Args:
            name: missing argument

        Returns:
            valid member of enum
        """
        for member in cls:
            if member.value.casefold() == name.casefold():
                return member


class FiwareLDHeader(BaseModel):
    """
    Define entity service paths which are supported by the NGSI
    Context Brokers to support hierarchical scopes:
    https://fiware-orion.readthedocs.io/en/master/user/service_path/index.html
    """

    model_config = ConfigDict(populate_by_name=True, validate_assignment=True)
    ngsild_tenant: str = Field(
        alias="NGSILD-Tenant",
        default=None,
        max_length=50,
        description="Alias to the Fiware service to used for multitenancy",
        pattern=r"\w*$",
    )
    link_header: str = Field(
        alias="Link",
        default=f"<{core_context}>; "
        'rel="http://www.w3.org/ns/json-ld#context"; '
        'type="application/ld+json"',
        description="Fiware service used for multi-tenancy",
        pattern=r"\w*$",
    )
    # @computed_field
    # def Link(self) -> str:
    #     link_header = f'<{self.context}>; ' \
    #                   'rel="http://www.w3.org/ns/json-ld#context"; ' \
    #                   'type="application/ld+json"'
    #     return link_header

    def set_context(self, context: str):
        self.link_header = f'<{context}>; rel="http://www.w3.org/ns/json-ld#context"; type="application/ld+json"'



"""
HTTP clients for FIWARE's NGSIv2 APIs
"""

from .cb import ContextBrokerClient
from .iota import IoTAClient
from .quantumleap import QuantumLeapClient
from .client import HttpClient, HttpClientConfig



"""Vocabulary Models for Ontology Entities"""

from enum import Enum
from pydantic import BaseModel, Field
from typing import List, TYPE_CHECKING, Dict, Union, Set, Any

from .source import DependencyStatement

if TYPE_CHECKING:
    from . import (
        CombinedObjectRelation,
        CombinedDataRelation,
        CombinedRelation,
        Relation,
        Vocabulary,
        Source,
    )


class Entity(BaseModel):
    """
    Representing an OWL Entity (Class, Datatype, DataProperty, ObjectProperty,
                                Individual)

    An Entity is characterised by a unique IRI and originates from a source

    An Entity needs a unique Label (displayname) as it is used in FIWARE as
    field key. The user can overwrite the given
    label
    """

    iri: str = Field(description="Unique Internationalized Resource Identifier")
    label: str = Field(
        default="",
        description="Label (displayname) extracted from source file "
        "(multiple Entities could have the same label)",
    )
    user_set_label: Any = Field(
        default="",
        description="Given by user and overwrites 'label'."
        " Needed to make labels unique",
    )
    comment: str = Field(
        default="", description="Comment extracted from the ontology/source"
    )
    source_ids: Set[str] = Field(
        default=set(), description="IDs of the sources that influenced this class"
    )
    predefined: bool = Field(
        default=False,
        description="Stats if the entity is not extracted from a source, "
        "but predefined in the program (Standard Datatypes)",
    )

    def get_label(self) -> str:
        """Get the label for the entity.
        If the user has set a label it is returned, else the label extracted
        from the source

        Returns:
             str
        """
        if not self.user_set_label == "":
            return self.user_set_label

        return self.get_original_label()

    def set_label(self, label: str):
        """Change the display label of the entity

        Args:
            label (str): Label that the label should have
        """
        self.user_set_label = label

    def get_ontology_iri(self) -> str:
        """Get the IRI of the ontology that this entity belongs to
        (extracted from IRI)

        Returns:
            str
        """
        index = self.iri.find("#")
        return self.iri[:index]

    def get_source_names(self, vocabulary: "Vocabulary") -> List[str]:
        """Get the names of all the sources

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            str
        """
        names = [vocabulary.get_source(id).get_name() for id in self.source_ids]

        return names

    def get_sources(self, vocabulary: "Vocabulary") -> List["Source"]:
        """Get all the source objects that influenced this entity.
        The sources are sorted according to their names

        Args:
           vocabulary (Vocabulary): Vocabulary of the project

        Returns:
           str
        """

        sources = [vocabulary.get_source(id) for id in self.source_ids]

        sources.sort(key=lambda x: x.source_name, reverse=False)
        return sources

    def _lists_are_identical(self, a: List, b: List) -> bool:
        """Methode to test if to lists contain the same entries

        Args:
            a (List): first list
            b (List): second list
        Returns:
            bool
        """
        return len(set(a).intersection(b)) == len(set(a)) and len(a) == len(b)

    def is_renamed(self) -> bool:
        """Check if the entity was renamed by the user

        Returns:
            bool
        """
        return not self.user_set_label == ""

    def get_original_label(self) -> str:
        """Get label as defined in the source
        It can be that the label is empty, then extract the label from the iri

        Returns:
            str
        """
        if not self.label == "":
            return self.label

        index = self.iri.find("#") + 1
        return self.iri[index:]


class Class(Entity):
    """
    Representation of OWL:CLASS

    A class has a set of relations that are combined into CombinedRelations

    Instances are instantiations of a class

    A class can represent Devices, Agents, None or both
    """

    # The objects whose ids/iris are listed here can be looked up in the
    # vocabulary of this class
    child_class_iris: List[str] = Field(
        default=[], description="All class_iris of classes that inherit from this class"
    )
    ancestor_class_iris: List[str] = Field(
        default=[],
        description="All class_iris of classes from which this class inherits",
    )
    parent_class_iris: List[str] = Field(
        default=[],
        description="All class_iris of classes that are direct parents of this "
        "class",
    )
    relation_ids: List[str] = Field(
        default=[], description="All ids of relations defined for this class"
    )
    combined_object_relation_ids: List[str] = Field(
        default=[],
        description="All combined_object_relations ids defined for this class",
    )
    combined_data_relation_ids: List[str] = Field(
        default=[], description="All combined_data_relations ids defined for this class"
    )

    def get_relation_ids(self) -> List[str]:
        """Get all ids of relations belonging to this class

        Returns:
            List[str]
        """
        return self.relation_ids

    def get_relations(self, vocabulary: "Vocabulary") -> List["Relation"]:
        """Get all relations belonging to this class

        Args:
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            List[Relation]
        """
        result = []
        for id in self.relation_ids:
            result.append(vocabulary.get_relation_by_id(id))

        return result

    def get_combined_object_relations(
        self, vocabulary: "Vocabulary"
    ) -> List["CombinedObjectRelation"]:
        """Get all combined object relations belonging to this class

        Args:
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            List[CombinedObjectRelation]
        """

        result = []
        for id in self.combined_object_relation_ids:
            result.append(vocabulary.get_combined_object_relation_by_id(id))

        return result

    def get_combined_data_relations(
        self, vocabulary: "Vocabulary"
    ) -> List["CombinedDataRelation"]:
        """Get all combined data relations belonging to this class

        Args:
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            List[CombinedDataRelation]
        """

        result = []
        for id in self.combined_data_relation_ids:
            result.append(vocabulary.get_combined_data_relation_by_id(id))

        return result

    def get_combined_relations(
        self, vocabulary: "Vocabulary"
    ) -> List["CombinedRelation"]:
        """Get all combined relations belonging to this class

        Args:
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            List[CombinedRelation]
        """

        result = self.get_combined_object_relations(vocabulary)
        result.extend(self.get_combined_data_relations(vocabulary))
        return result

    def is_child_of_all_classes(self, target_list: List[str]) -> bool:
        """Tests if this class is a child class for each of the given classes

        Args:
            target_list (List[str]): List of ancestor class_iris

        Returns:
            bool
        """

        for target in target_list:
            if not target == self.iri:
                if target not in self.ancestor_class_iris:
                    return False
        return True

    def get_combined_object_relation_with_property_iri(
        self, obj_prop_iri: str, vocabulary: "Vocabulary"
    ) -> "CombinedObjectRelation":
        """
        Get the CombinedObjectRelation of this class that combines the
        relations of the given ObjectProperty

        Args:
            obj_prop_iri (str): Iri of the ObjectProperty
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            CombinedObjectRelation
        """
        for cor in self.get_combined_object_relations(vocabulary):
            if cor.property_iri == obj_prop_iri:
                return cor
        return None

    def get_combined_data_relation_with_property_iri(self, property_iri, vocabulary):
        """
        Get the CombinedDataRelation of this class that combines the
        relations of the given DataProperty

        Args:
            property_iri (str): Iri of the DataProperty
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            CombinedDataRelation
        """
        for cdr in self.get_combined_data_relations(vocabulary):
            if cdr.property_iri == property_iri:
                return cdr
        return None

    def get_combined_relation_with_property_iri(
        self, property_iri, vocabulary
    ) -> Union["CombinedRelation", None]:
        """
        Get the CombinedRelation of this class that combines the relations
        of the given Property

        If possible use the more specific access functions to save runtime.

        Args:
            property_iri (str): Iri of the Property
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            CombinedRelation, None if iri is unknown
        """
        for cdr in self.get_combined_data_relations(vocabulary):
            if cdr.property_iri == property_iri:
                return cdr
        for cor in self.get_combined_object_relations(vocabulary):
            if cor.property_iri == property_iri:
                return cor
        return None

    def get_ancestor_classes(self, vocabulary: "Vocabulary") -> List["Class"]:
        """Get all ancestor classes of this class

        Args:
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            List[Class]
        """
        ancestors = []
        for ancestor_iri in self.ancestor_class_iris:
            ancestors.append(vocabulary.get_class_by_iri(ancestor_iri))
        return ancestors

    def get_parent_classes(
        self, vocabulary: "Vocabulary", remove_redundancy: bool = False
    ) -> List["Class"]:
        """Get all parent classes of this class

        Args:
            vocabulary (Vocabulary): Vocabulary of this project
            remove_redundancy (bool): if true the parents that are child of
                other parents are not included

        Returns:
            List[Class]
        """
        parents = []

        for parent_iri in self.parent_class_iris:
            parents.append(vocabulary.get_class_by_iri(parent_iri))

        if remove_redundancy:
            child_iris = set()
            for parent in parents:
                child_iris.update(parent.child_class_iris)
            for parent in parents:
                if parent.iri in child_iris:
                    parents.remove(parent)

        return parents

    def treat_dependency_statements(
        self, vocabulary: "Vocabulary"
    ) -> List[DependencyStatement]:
        """
        Purge and list all pointers/iris that are not contained in
        the vocabulary

        Args:
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            List[Dict[str, str]]: List of purged statements dicts with keys:
            Parent Class, class, dependency, fulfilled
        """

        statements = []
        # parent classes:
        parents_to_purge = []
        for parent_iri in self.parent_class_iris:
            found = parent_iri in vocabulary.classes
            statements.append(
                DependencyStatement(
                    type="Parent Class",
                    class_iri=self.iri,
                    dependency_iri=parent_iri,
                    fulfilled=found,
                )
            )
            if not found:
                parents_to_purge.append(parent_iri)
        for iri in parents_to_purge:
            self.parent_class_iris.remove(iri)

        # relations
        relation_ids_to_purge = set()
        for relation in self.get_relations(vocabulary):

            relation_statements = relation.get_dependency_statements(
                vocabulary, self.get_ontology_iri(), self.iri
            )
            for statement in relation_statements:
                if statement.fulfilled == False:
                    relation_ids_to_purge.add(relation.id)
            statements.extend(relation_statements)

        for id in relation_ids_to_purge:
            self.relation_ids.remove(id)
            del vocabulary.relations[id]

        return statements

    def get_next_combined_relation_id(
        self, current_cr_id: str, object_relations: bool
    ) -> str:
        """Get the alphabetically(Property label) next CombinedRelation.

        If no CR is after the given one, the first is returned

        Args:
            current_cr_id (str): ID of the CombinedRelation of which the next
                should be found
            object_relations (bool):
                True if Searching for  CombinedObjectRelations

        Returns:
            str: ID of next CR
        """
        list_ = self.combined_data_relation_ids
        if object_relations:
            list_ = self.combined_object_relation_ids

        current_index = list_.index(current_cr_id)
        res_index = current_index + 1
        if res_index >= len(list_):
            res_index = 0
        return list_[res_index]

    def get_previous_combined_relation_id(
        self, current_cr_id: str, object_relations: bool
    ) -> str:
        """Get the alphabetically(Property label) previous CombinedRelation.

        If no CR is before the given one, the last is returned

        Args:
            current_cr_id (str): ID of the CombinedRelation of which the
                previous should be found
            object_relations (bool): True if Searching for
                CombinedObjectRelations

        Returns:
            str: ID of previous CR
        """

        list_ = self.combined_data_relation_ids
        if object_relations:
            list_ = self.combined_object_relation_ids

        current_index = list_.index(current_cr_id)
        res_index = current_index - 1
        if res_index < 0:
            res_index = len(list_) - 1
        return list_[res_index]

    def is_logically_equivalent_to(
        self, class_: "Class", vocabulary: "Vocabulary", old_vocabulary: "Vocabulary"
    ) -> bool:
        """Test if a class is logically equivalent in two vocabularies.

        Args:
            class_ (Class): Class to be tested against, from the old_vocabulary
            vocabulary (Vocabulary): New project vocabulary
            old_vocabulary (Vocabulary): Old project vocabulary

        Returns:
            bool
        """

        # test if parent classes are identical
        if not self._lists_are_identical(
            class_.parent_class_iris, self.parent_class_iris
        ):
            return False

        # test if combined object relation ids are identical
        if not self._lists_are_identical(
            class_.combined_object_relation_ids, self.combined_object_relation_ids
        ):
            return False

        # test if combined data  relation ids are identical
        if not self._lists_are_identical(
            class_.combined_data_relation_ids, self.combined_data_relation_ids
        ):
            return False

        # test if combined relations are identical
        for cr in self.get_combined_relations(vocabulary):
            old_cr = old_vocabulary.get_combined_relation_by_id(cr.id)

            relation_strings = []
            for relation in cr.get_relations(vocabulary):
                relation_strings.append(relation.to_string(vocabulary))

            old_relation_strings = []
            for old_relation in old_cr.get_relations(old_vocabulary):
                old_relation_strings.append(old_relation.to_string(vocabulary))

            if not self._lists_are_identical(relation_strings, old_relation_strings):

                return False

        return True

    def is_iot_class(self, vocabulary: "Vocabulary") -> bool:
        """
        A class is an iot/device class if it contains one CDR, where the
        relation is marked as a device relation: DeviceAttribute/Command

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            bool
        """

        for cdr_id in self.combined_data_relation_ids:
            cdr = vocabulary.get_combined_data_relation_by_id(cdr_id)
            prop = vocabulary.get_data_property(cdr.property_iri)
            if not prop.field_type == DataFieldType.simple:
                return True
        return False


class DatatypeType(str, Enum):
    """
    Types of a Datatype
    """

    string = "string"
    number = "number"
    date = "date"
    enum = "enum"


class DatatypeFields(BaseModel):
    """Key Fields describing a Datatype"""

    type: DatatypeType = Field(
        default=DatatypeType.string, description="Type of the datatype"
    )
    number_has_range: Any = Field(
        default=False, description="If Type==Number: Does the datatype define a range"
    )
    number_range_min: Union[int, str] = Field(
        default="/",
        description="If Type==Number: Min value of the datatype range, "
        "if a range is defined",
    )
    number_range_max: Union[int, str] = Field(
        default="/",
        description="If Type==Number: Max value of the datatype range, "
        "if a range is defined",
    )
    number_decimal_allowed: bool = Field(
        default=False, description="If Type==Number: Are decimal numbers allowed?"
    )
    forbidden_chars: List[str] = Field(
        default=[], description="If Type==String: Blacklisted chars"
    )
    allowed_chars: List[str] = Field(
        default=[], description="If Type==String: Whitelisted chars"
    )
    enum_values: List[str] = Field(default=[], description="If Type==Enum: Enum values")


class Datatype(Entity, DatatypeFields):
    """
    Represents OWL:Datatype

    A Datatype is the target of a DataRelation. The Datatype stats a set of
    values that are valid.
    This can be an ENUM, a number range, or a check for black/whitelisted chars

    In the Parsing PostProcesseor predefined datatype_catalogue are added to the
    vocabulary
    """

    def export(self) -> Dict[str, str]:
        """Export datatype as dict

        Returns:
            Dict[str,str]
        """
        res = self.model_dump(
            include={
                "type",
                "number_has_range",
                "number_range_min",
                "number_range_max",
                "number_decimal_allowed",
                "forbidden_chars",
                "allowed_chars",
                "enum_values",
            },
            exclude_defaults=True,
        )
        res["type"] = self.type.value
        return res

    def value_is_valid(self, value: str) -> bool:
        """Test if value is valid for this datatype.
        Numbers are also given as strings

        Args:
            value (str): value to be tested

        Returns:
            bool
        """

        if self.type == DatatypeType.string:
            if len(self.allowed_chars) > 0:
                # if allowed chars is empty all chars are allowed
                for char in value:
                    if char not in self.allowed_chars:
                        return False
            for char in self.forbidden_chars:
                if char in value:
                    return False
            return True

        if self.type == DatatypeType.number:

            if self.number_decimal_allowed:
                try:
                    number = float(value)
                except:
                    return False
            else:
                try:
                    number = int(value)
                except:
                    return False

            if not self.number_range_min == "/":
                if number < self.number_range_min:
                    return False
            if not self.number_range_max == "/":
                if number > self.number_range_max:
                    return False

            return True

        if self.type == DatatypeType.enum:
            return value in self.enum_values

        if self.type == DatatypeType.date:
            try:
                from dateutil.parser import parse

                parse(value, fuzzy=False)
                return True

            except ValueError:
                return False

        return True

    def is_logically_equivalent_to(
        self,
        datatype: "Datatype",
        vocabulary: "Vocabulary",
        old_vocabulary: "Vocabulary",
    ) -> bool:
        """Test if this datatype is logically equivalent to the given datatype

        Args:
            datatype (Datatype): Datatype to compare against
            vocabulary (Vocabulary): Not used, but needed to keep signature the
                same as other entities
            old_vocabulary (Vocabulary): Not used, but needed to keep signature
                the same as other entities
        Returns:
            bool
        """

        if not self.type == datatype.type:
            return False
        if not self.number_has_range == datatype.number_has_range:
            return False
        if not self.enum_values == datatype.enum_values:
            return False

        return True


class Individual(Entity):
    """
    Represents OWL:Individual

    An individual is a predefined "instance" of a class
    But they are here only used as values for Relations

    They are not instances, no value can be assigned to them, they are no
    agents or devices
    """

    parent_class_iris: List[str] = Field(
        default=[],
        description="List of all parent class iris, "
        "an individual can have multiple parents",
    )

    def to_string(self) -> str:
        """Get a string representation of the Individual

        Returns:
            str
        """
        return "(Individual)" + self.get_label()

    def get_ancestor_iris(self, vocabulary: "Vocabulary") -> List[str]:
        """Get all iris of ancestor classes

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            List[str]
        """
        ancestor_iris = set()
        for parent_iri in self.parent_class_iris:
            ancestor_iris.add(parent_iri)
            ancestor_iris.update(
                vocabulary.get_class_by_iri(parent_iri).ancestor_class_iris
            )

        return list(ancestor_iris)

    def get_parent_classes(self, vocabulary: "Vocabulary") -> List["Class"]:
        """Get all parent class objects

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            List[Class]
        """
        parents = []
        for parent_iri in self.parent_class_iris:
            parents.append(vocabulary.get_class_by_iri(parent_iri))
        return parents

    def is_logically_equivalent_to(
        self,
        individual: "Individual",
        vocabulary: "Vocabulary",
        old_vocabulary: "Vocabulary",
    ) -> bool:
        """Test if this individal is logically equivalent in two vocabularies.

        Args:
            individual (Individual): Individual to be tested against, from the
                old vocabulary
            vocabulary (Vocabulary): New project vocabulary, not used but needed
                to keep signature the same
            old_vocabulary (Vocabulary): Old project vocabulary, not used but
                needed to keep signature the same

        Returns:
            bool
        """

        if not self._lists_are_identical(
            self.parent_class_iris, individual.parent_class_iris
        ):
            return False
        return True

    def treat_dependency_statements(
        self, vocabulary: "Vocabulary"
    ) -> List[DependencyStatement]:
        """Purge and list all pointers/iris that are not contained in the
        vocabulary

        Args:
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            List[Dict[str, str]]: List of purged statements dicts with keys:
            Parent Class, class, dependency, fulfilled
        """
        statements = []

        for parent_iri in self.parent_class_iris:
            found = parent_iri in vocabulary.classes
            statements.append(
                DependencyStatement(
                    type="Parent Class",
                    class_iri=self.iri,
                    dependency_iri=parent_iri,
                    fulfilled=found,
                )
            )

            if not found:
                self.parent_class_iris.remove(parent_iri)

        return statements


class DataFieldType(str, Enum):
    """Type of the field that represents the DataProperty"""

    command = "command"
    device_attribute = "device_attribute"
    simple = "simple"


class DataProperty(Entity):
    """
    Representation of OWL:DataProperty
    """

    field_type: DataFieldType = Field(
        default=DataFieldType.simple,
        description="Type of the dataproperty; set by the user while "
        "configuring the vocabulary",
    )


class ObjectProperty(Entity):
    """
    Representation of OWL:ObjectProperty
    """

    inverse_property_iris: Set[str] = Field(
        default=set(),
        description="List of property iris that are inverse:Of; "
        "If an instance i2 is added in an instance i1 "
        "for this property. Then i1 is added to i2 under the"
        " inverseProperty (if the class has that property)",
    )

    def add_inverse_property_iri(self, iri: str):
        """Add an inverse property

        Args:
            iri (str): Iri of the inverse objectProperty

        Returns:
            None
        """
        self.inverse_property_iris.add(iri)

    def is_logically_equivalent_to(
        self,
        object_property: "ObjectProperty",
        vocabulary: "Vocabulary",
        old_vocabulary: "Vocabulary",
    ) -> bool:
        """Test if this Property in the new_vocabulary is logically equivalent
        to the object_property in the old_vocabulary

        Args:
            object_property (ObjectProperty): ObjectProperty to be tested
                against, from the old vocabulary
            vocabulary (Vocabulary): New project vocabulary, not used but
                needed to keep signature the same
            old_vocabulary (Vocabulary): Old project vocabulary, not used but
                needed to keep signature the same

        Returns:
            bool
        """
        if not self.inverse_property_iris == object_property.inverse_property_iris:
            return False

        return True



"""
This module contains NGSIv2 models for context subscription in the context
broker.
"""

from typing import Any, List, Dict, Union, Optional
from datetime import datetime
from aenum import Enum
from pydantic import (
    field_validator,
    model_validator,
    ConfigDict,
    BaseModel,
    conint,
    Field,
    Json,
)
from .base import AttrsFormat, EntityPattern, Http, Status, Expression
from filip.utils.validators import validate_mqtt_url, validate_mqtt_topic
from filip.models.ngsi_v2.context import ContextEntity
from filip.models.ngsi_v2.base import EntityPattern, Expression, DataType
from filip.custom_types import AnyMqttUrl
import warnings

# The pydantic models still have a .json() function, but this method is deprecated.
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    message='Field name "json" shadows an attribute in parent "Http"',
)
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    message='Field name "json" shadows an attribute in parent "Mqtt"',
)


class NgsiPayloadAttr(BaseModel):
    """
    Model for NGSI V2 type payload in httpCustom/mqttCustom notifications.
    The difference between this model and the usual BaseValueAttribute model is that
    a metadata field is not allowed.
    In the absence of type/value in some attribute field, one should resort to partial
    representations ( as specified in the orion api manual), done by the BaseValueAttr.
    model.
    """

    model_config = ConfigDict(extra="forbid")
    type: Union[DataType, str] = Field(
        default=DataType.TEXT,
        description="The attribute type represents the NGSI value type of the "
        "attribute value. Note that FIWARE NGSI has its own type "
        "system for attribute values, so NGSI value types are not "
        "the same as JSON types. Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
    )
    value: Optional[Any] = Field(
        default=None, title="Attribute value", description="the actual data"
    )


class NgsiPayload(BaseModel):
    """
    Model for NGSI V2 type payload in httpCustom/mqttCustom notifications.
    Differences between this model and the usual Context entity models include:
        - id and type are not mandatory
        - an attribute metadata field is not allowed
    """

    model_config = ConfigDict(extra="allow", validate_default=True)
    id: Optional[str] = Field(default=None, max_length=256, min_length=1, frozen=True)
    type: Optional[Union[str, Enum]] = Field(
        default=None,
        max_length=256,
        min_length=1,
        frozen=True,
    )

    @model_validator(mode="after")
    def validate_notification_attrs(self):
        for v in self.model_dump(exclude={"id", "type"}).values():
            assert isinstance(NgsiPayloadAttr.model_validate(v), NgsiPayloadAttr)
        return self


class Message(BaseModel):
    """
    Model for a notification message, when sent to other NGSIv2-APIs
    """

    subscriptionId: Optional[str] = Field(
        default=None,
        description="Id of the subscription the notification comes from",
    )
    data: List[ContextEntity] = Field(
        description="is an array with the notification data itself which "
        "includes the entity and all concerned attributes. Each "
        "element in the array corresponds to a different entity. "
        "By default, the entities are represented in normalized "
        "mode. However, using the attrsFormat modifier, a "
        "simplified representation mode can be requested."
    )


class HttpMethods(str, Enum):
    _init_ = "value __doc__"

    POST = "POST", "Post Method"
    PUT = "PUT", "Put Method"
    PATCH = "PATCH", "Patch Method"


class HttpCustom(Http):
    """
    Model for custom notification patterns sent via HTTP
    """

    headers: Optional[Dict[str, Union[str, Json]]] = Field(
        default=None,
        description="a key-map of HTTP headers that are included in "
        "notification messages.",
    )
    qs: Optional[Dict[str, Union[str, Json]]] = Field(
        default=None,
        description="a key-map of URL query parameters that are included in "
        "notification messages.",
    )
    method: str = Field(
        default=HttpMethods.POST,
        description="the method to use when sending the notification "
        "(default is POST). Only valid HTTP methods are allowed. "
        "On specifying an invalid HTTP method, a 400 Bad Request "
        "error is returned.",
    )
    payload: Optional[str] = Field(
        default=None,
        description="the payload to be used in notifications. If omitted, the "
        'default payload (see "Notification Messages" sections) '
        "is used.",
    )
    json: Optional[Dict[str, Union[str, Json]]] = Field(
        default=None,
        description="get a json as notification. If omitted, the default"
        'payload (see "Notification Messages" sections) is used.',
    )
    ngsi: Optional[NgsiPayload] = Field(
        default=None,
        description="get an NGSI-v2 normalized entity as notification.If omitted, "
        'the default payload (see "Notification Messages" sections) is used.',
    )
    timeout: Optional[int] = Field(
        default=None,
        description="Maximum time (in milliseconds) the subscription waits for the "
        "response. The maximum value allowed for this parameter is 1800000 "
        "(30 minutes). If timeout is defined to 0 or omitted, then the value "
        "passed as -httpTimeout CLI parameter is used. See section in the "
        "'Command line options' for more details.",
    )

    @model_validator(mode="after")
    def validate_notification_payloads(self):
        fields = [self.payload, self.json, self.ngsi]
        filled_fields = [field for field in fields if field is not None]

        if len(filled_fields) > 1:
            raise ValueError(
                "Only one of payload, json or ngsi fields accepted at the "
                "same time in httpCustom."
            )

        return self


class Mqtt(BaseModel):
    """
    Model for notifications sent via MQTT
    https://fiware-orion.readthedocs.io/en/3.8.0/user/mqtt_notifications/index.html
    """

    url: Union[AnyMqttUrl, str] = Field(
        description="to specify the MQTT broker endpoint to use. URL must "
        "start with mqtt:// and never contains a path (i.e. it "
        "only includes host and port)"
    )
    topic: str = Field(
        description="to specify the MQTT topic to use",
    )
    valid_type = field_validator("topic")(validate_mqtt_topic)
    qos: Optional[int] = Field(
        default=0,
        description="to specify the MQTT QoS value to use in the "
        "notifications associated to the subscription (0, 1 or 2). "
        "This is an optional field, if omitted then QoS 0 is used.",
        ge=0,
        le=2,
    )
    user: Optional[str] = Field(default=None, description="username if required")
    passwd: Optional[str] = Field(default=None, description="password if required")

    @field_validator("url")
    @classmethod
    def check_url(cls, value):
        """
        Check if url has a valid structure
        Args:
            value: url to validate
        Returns:
            validated url
        """
        return validate_mqtt_url(url=value)


class MqttCustom(Mqtt):
    """
    Model for custom notification patterns sent via MQTT
    https://fiware-orion.readthedocs.io/en/3.8.0/user/mqtt_notifications/index.html
    """

    payload: Optional[str] = Field(
        default=None,
        description="the payload to be used in notifications. If omitted, the "
        'default payload (see "Notification Messages" sections) '
        "is used.",
    )
    json: Optional[Dict[str, Any]] = Field(
        default=None,
        description="get a json as notification. If omitted, the default"
        'payload (see "Notification Messages" sections) is used.',
    )
    ngsi: Optional[NgsiPayload] = Field(
        default=None,
        description="get an NGSI-v2 normalized entity as notification.If omitted, "
        'the default payload (see "Notification Messages" sections) is used.',
    )

    @model_validator(mode="after")
    def validate_payload_type(self):
        assert (
            len(
                [
                    v
                    for k, v in self.model_dump().items()
                    if ((v is not None) and (k in ["payload", "ngsi", "json"]))
                ]
            )
            <= 1
        )
        return self


class Notification(BaseModel):
    """
    If the notification attributes are left empty, all attributes will be
    included in the notifications. Otherwise, only the specified ones will
    be included.
    """

    model_config = ConfigDict(validate_assignment=True)
    timesSent: Optional[Any] = Field(
        default=None,
        description="Not editable, only present in GET operations. "
        "Number of notifications sent due to this subscription.",
    )
    http: Optional[Http] = Field(
        default=None,
        description="It is used to convey parameters for notifications "
        "delivered through the HTTP protocol. Cannot be used "
        'together with "httpCustom, mqtt, mqttCustom"',
    )
    httpCustom: Optional[HttpCustom] = Field(
        default=None,
        description="It is used to convey parameters for notifications "
        "delivered through the HTTP protocol. Cannot be used "
        'together with "http"',
    )
    mqtt: Optional[Mqtt] = Field(
        default=None,
        description="It is used to convey parameters for notifications "
        "delivered through the MQTT protocol. Cannot be used "
        'together with "http, httpCustom, mqttCustom"',
    )
    mqttCustom: Optional[MqttCustom] = Field(
        default=None,
        description="It is used to convey parameters for notifications "
        "delivered through the MQTT protocol. Cannot be used "
        'together with "http, httpCustom, mqtt"',
    )
    attrs: Optional[List[str]] = Field(
        default=None,
        description="List of attributes to be included in notification "
        "messages. It also defines the order in which attributes "
        "must appear in notifications when attrsFormat value is "
        'used (see "Notification Messages" section). An empty list '
        "means that all attributes are to be included in "
        'notifications. See "Filtering out attributes and '
        'metadata" section for more detail.',
    )
    exceptAttrs: Optional[List[str]] = Field(
        default=None,
        description="List of attributes to be excluded from the notification "
        "message, i.e. a notification message includes all entity "
        "attributes except the ones listed in this field.",
    )
    attrsFormat: Optional[AttrsFormat] = Field(
        default=AttrsFormat.NORMALIZED,
        description="specifies how the entities are represented in "
        "notifications. Accepted values are normalized (default), "
        "keyValues or values. If attrsFormat takes any value "
        "different than those, an error is raised. See detail in "
        '"Notification Messages" section.',
    )
    metadata: Optional[Any] = Field(
        default=None,
        description="List of metadata to be included in notification messages. "
        'See "Filtering out attributes and metadata" section for '
        "more detail.",
    )
    onlyChangedAttrs: Optional[bool] = Field(
        default=False,
        description="Only supported by Orion Context Broker!"
        "If set to true then notifications associated to the "
        "subscription include only attributes that changed in the "
        "triggering update request, in combination with the attrs "
        "or exceptAttrs field. For instance, if attrs is "
        "[A=1, B=2, C=3] and A=0 is updated. In case "
        "onlyChangedAttrs=false, CB notifies [A=0, B=2, C=3]."
        "In case onlyChangedAttrs=true, CB notifies "
        "[A=0, B=null, C=null]. This ",
    )
    lastSuccess: Optional[Any] = Field(
        default=None,
        description="Not editable, only present in GET operations. "
        "A Timestamp in ISO8601 format for last successful "
        "notification. Not present if subscription has never "
        "had a successful notification.",
    )
    lastSuccessCode: Optional[Any] = Field(
        default=None,
        description="Not editable, only present in GET operations. "
        "HTTP status code of the last successful notification.",
    )
    lastFailure: Optional[Any] = Field(
        default=None,
        description="Not editable, only present in GET operations"
        "Last failure timestamp in ISO8601 format. Not present if "
        "subscription has never had a problem with notifications.",
    )
    lastFailureReason: Optional[Any] = Field(
        default=None,
        description="Not editable, only present in GET operations. "
        "A string with the reason for the last failure. ",
    )
    covered: Optional[bool] = Field(
        default=False,
        description="A flag to decide whether to include not existing attribute in "
        "notifications. It can be useful for those notification endpoints "
        "that are not flexible enough for a variable set of attributes and "
        "needs always the same set of incoming attributes in every received"
        " notification "
        "https://fiware-orion.readthedocs.io/en/master/orion-api.html#covered-subscriptions",
    )

    @model_validator(mode="after")
    def validate_http(self):
        if self.httpCustom is not None:
            assert self.http is None
        return self

    @model_validator(mode="after")
    def validate_attr(self):
        if self.exceptAttrs is not None:
            assert self.attrs is None
        return self

    @model_validator(mode="after")
    def validate_endpoints(self):
        if self.http is not None:
            assert all(
                (
                    v is None
                    for k, v in self.model_dump().items()
                    if k in ["httpCustom", "mqtt", "mqttCustom"]
                )
            )
        elif self.httpCustom is not None:
            assert all(
                (
                    v is None
                    for k, v in self.model_dump().items()
                    if k in ["http", "mqtt", "mqttCustom"]
                )
            )
        elif self.mqtt is not None:
            assert all(
                (
                    v is None
                    for k, v in self.model_dump().items()
                    if k in ["http", "httpCustom", "mqttCustom"]
                )
            )
        else:
            assert all(
                (
                    v is None
                    for k, v in self.model_dump().items()
                    if k in ["http", "httpCustom", "mqtt"]
                )
            )
        return self

    @model_validator(mode="after")
    def validate_covered_attrs(self):
        if self.covered is True:
            if isinstance(self.attrs, list) and len(self.attrs) > 0:
                return self
            else:
                raise ValueError("Covered notification need an explicit list of attrs.")
        return self


class Response(Notification):
    """
    Server response model for notifications
    """

    timesSent: int = Field(
        description="(not editable, only present in GET operations): "
        "Number of notifications sent due to this subscription."
    )
    lastNotification: datetime = Field(
        description="(not editable, only present in GET operations): "
        "Last notification timestamp in ISO8601 format."
    )
    lastFailure: Optional[datetime] = Field(
        default=None,
        description="(not editable, only present in GET operations): "
        "Last failure timestamp in ISO8601 format. Not present if "
        "subscription has never had a problem with notifications.",
    )
    lastSuccess: Optional[datetime] = Field(
        default=None,
        description="(not editable, only present in GET operations): "
        "Timestamp in ISO8601 format for last successful "
        "notification. Not present if subscription has never "
        "had a successful notification.",
    )


class Condition(BaseModel):
    """
    Notification rules are as follow:
    If attrs and expression are used, a notification is sent whenever one of
    the attributes in the attrs list changes and at the same time expression
    matches.
    If attrs is used and expression is not used, a notification is sent
    whenever any of the attributes in the attrs list changes.
    If attrs is not used and expression is used, a notification is sent
    whenever any of the attributes of the entity changes and at the same time
    expression matches.
    If neither attrs nor expression are used, a notification is sent whenever
    any of the attributes of the entity changes.
    alterationTypes: for more information about this field, see
    https://github.com/telefonicaid/fiware-orion/blob/3.8.0/doc/manuals/orion-api.md#subscriptions-based-in-alteration-type

    """

    attrs: Optional[Union[str, List[str]]] = Field(
        default=None, description="array of attribute names"
    )
    expression: Optional[Union[str, Expression]] = Field(
        default=None,
        description="an expression composed of q, mq, georel, geometry and "
        'coords (see "List entities" operation above about this '
        "field).",
    )
    alterationTypes: Optional[List[str]] = Field(
        default=None, description="list of alteration types triggering the subscription"
    )

    @field_validator("attrs")
    def check_attrs(cls, v):
        if isinstance(v, list):
            return v
        elif isinstance(v, str):
            return [v]
        else:
            raise TypeError()

    @field_validator("alterationTypes")
    def check_alteration_types(cls, v):
        allowed_types = {"entityCreate", "entityDelete", "entityUpdate", "entityChange"}

        if v is None:
            return None
        elif isinstance(v, list):
            for item in v:
                if item not in allowed_types:
                    raise ValueError(
                        f"{item} is not a valid alterationType"
                        f" allowed values are {allowed_types}"
                    )
            return v
        else:
            raise ValueError("alterationTypes must be a list of strings")


class Subject(BaseModel):
    """
    Model for subscription subject
    """

    entities: List[EntityPattern] = Field(
        description="A list of objects, each one composed of by an Entity " "Object:"
    )
    condition: Optional[Condition] = Field(
        default=None,
    )


class Subscription(BaseModel):
    """
    Subscription payload validations
    https://fiware-orion.readthedocs.io/en/master/user/ngsiv2_implementation_notes/index.html#subscription-payload-validations
    """

    model_config = ConfigDict(validate_assignment=True)

    id: Optional[str] = Field(
        default=None,
        description="Subscription unique identifier. Automatically created at "
        "creation time.",
    )
    description: Optional[str] = Field(
        default=None,
        description="A free text used by the client to describe the " "subscription.",
    )
    status: Optional[Status] = Field(
        default=Status.ACTIVE,
        description="Either active (for active subscriptions) or inactive "
        "(for inactive subscriptions). If this field is not "
        "provided at subscription creation time, new subscriptions "
        "are created with the active status, which can be changed"
        " by clients afterwards. For expired subscriptions, this "
        "attribute is set to expired (no matter if the client "
        "updates it to active/inactive). Also, for subscriptions "
        "experiencing problems with notifications, the status is "
        "set to failed. As soon as the notifications start working "
        "again, the status is changed back to active.",
    )
    subject: Subject = Field(
        description="An object that describes the subject of the subscription.",
        json_schema_extra={
            "example": {
                "entities": [{"idPattern": ".*", "type": "Room"}],
                "condition": {
                    "attrs": ["temperature"],
                    "expression": {"q": "temperature>40"},
                },
            }
        },
    )
    notification: Notification = Field(
        description="An object that describes the notification to send when "
        "the subscription is triggered.",
        json_schema_extra={
            "example": {
                "http": {"url": "http://localhost:1234"},
                "attrs": ["temperature", "humidity"],
            }
        },
    )
    expires: Optional[datetime] = Field(
        default=None,
        description="Subscription expiration date in ISO8601 format. "
        "Permanent subscriptions must omit this field.",
    )

    throttling: Optional[
        conint(
            strict=True,
            ge=0,
        )
    ] = Field(
        default=None,
        strict=True,
        description="Minimal period of time in seconds which "
        "must elapse between two consecutive notifications. "
        "It is optional.",
    )



"""
Base http client module
"""

import logging
from pydantic import AnyHttpUrl
from typing import Dict, ByteString, List, IO, Tuple, Union
import requests
from filip.models.base import FiwareHeader, FiwareLDHeader
from filip.utils import validate_http_url
from enum import Enum


class NgsiURLVersion(str, Enum):
    """
    URL part that defines the NGSI version for the API.
    """

    v2_url = "v2"
    ld_url = "ngsi-ld/v1"


class BaseHttpClient:
    """
    Base client for all derived api-clients.

    Args:
        session: request session object. This is required for reusing
            the same connection
        reuse_session (bool):
        fiware_header: Fiware header object required for multi tenancy
        **kwargs: Optional arguments that ``request`` takes.

    """

    def __init__(
        self,
        url: Union[AnyHttpUrl, str] = None,
        *,
        session: requests.Session = None,
        fiware_header: Union[Dict, FiwareHeader, FiwareLDHeader] = None,
        **kwargs,
    ):

        self.logger = logging.getLogger(name=f"{self.__class__.__name__}")
        self.logger.addHandler(logging.NullHandler())
        self.logger.debug("Creating %s", self.__class__.__name__)

        if url:
            self.logger.debug("Checking url style...")
            self.base_url = validate_http_url(url)

        if session:
            self.session = session
            self._external_session = True
        else:
            self.session = None
            self._headers = {}

        if not fiware_header:
            self.fiware_headers = FiwareHeader()
        else:
            self.fiware_headers = fiware_header

        self.headers.update(kwargs.pop("headers", {}))
        self.kwargs: Dict = kwargs

    # Context Manager Protocol
    def __enter__(self):
        if not self.session:
            self.session = requests.Session()
            self.headers.update(self.fiware_headers.model_dump(by_alias=True))
            self._external_session = False
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def fiware_headers(self) -> FiwareHeader:
        """
        Get fiware header

        Returns:
            FiwareHeader
        """
        return self._fiware_headers

    @fiware_headers.setter
    def fiware_headers(self, headers: Union[Dict, FiwareHeader]) -> None:
        """
        Sets new fiware header

        Args:
            headers (Dict, FiwareHeader): New headers either as FiwareHeader
                object or as dict.

        Example:
            {fiware-service: "MyService",
             fiware-servicepath: "/MyServicePath"}

        Returns:
            None
        """
        if isinstance(headers, FiwareHeader):
            self._fiware_headers = headers
        elif isinstance(headers, FiwareLDHeader):
            self._fiware_headers = headers
        elif isinstance(headers, dict):
            self._fiware_headers = FiwareHeader.model_validate(headers)
        elif isinstance(headers, str):
            self._fiware_headers = FiwareHeader.model_validate_json(headers)
        elif isinstance(headers, dict):
            self._fiware_headers = FiwareLDHeader.parse_obj(headers)
        elif isinstance(headers, str):
            self._fiware_headers = FiwareLDHeader.parse_raw(headers)
        else:
            raise TypeError(f"Invalid headers! {type(headers)}")
        self.headers.update(self.fiware_headers.model_dump(by_alias=True))

    @property
    def fiware_service(self) -> str:
        """
        Get current fiware service
        Returns:
            str
        """
        return self.fiware_headers.service

    @fiware_service.setter
    def fiware_service(self, service: str) -> None:
        """
        Set new fiware service
        Args:
            service:

        Returns:
            None
        """
        self._fiware_headers.service = service
        self.headers.update(self.fiware_headers.model_dump(by_alias=True))

    @property
    def fiware_service_path(self) -> str:
        """
        Get current fiware service path
        Returns:
            str
        """
        return self.fiware_headers.service_path

    @fiware_service_path.setter
    def fiware_service_path(self, service_path: str) -> None:
        """
        Set new fiware service path
        Args:
            service_path (str): New fiware service path. Must start with '/'

        Returns:
            None
        """
        self._fiware_headers.service_path = service_path
        self.headers.update(self.fiware_headers.model_dump(by_alias=True))

    @property
    def headers(self):
        """
        Return current session headers
        Returns:
            dict with headers
        """
        if self.session:
            return self.session.headers
        return self._headers

    # modification to requests api
    def get(
        self, url: str, params: Union[Dict, List[Tuple], ByteString] = None, **kwargs
    ) -> requests.Response:
        """
        Sends a GET request either using the provided session or the single
        session.

        Args:
            url (str): URL for the new :class:`Request` object.
            params (optional): (optional) Dictionary, list of tuples or bytes
                to send in the query string for the :class:`Request`.
            **kwargs: Optional arguments that ``request`` takes.

        Returns:
            requests.Response
        """

        kwargs.update({k: v for k, v in self.kwargs.items() if k not in kwargs.keys()})

        if self.session:
            return self.session.get(url=url, params=params, **kwargs)
        return requests.get(url=url, params=params, **kwargs)

    def options(self, url: str, **kwargs) -> requests.Response:
        """
        Sends an OPTIONS request either using the provided session or the
        single session.

        Args:
            url (str):
            **kwargs: Optional arguments that ``request`` takes.

        Returns:
            requests.Response
        """
        kwargs.update({k: v for k, v in self.kwargs.items() if k not in kwargs.keys()})

        if self.session:
            return self.session.options(url=url, **kwargs)
        return requests.options(url=url, **kwargs)

    def head(
        self, url: str, params: Union[Dict, List[Tuple], ByteString] = None, **kwargs
    ) -> requests.Response:
        """
        Sends a HEAD request either using the provided session or the
        single session.

        Args:
            url (str): URL for the new :class:`Request` object.
            params (optional): Dictionary, list of tuples or bytes
                to send in the query string for the :class:`Request`.
            **kwargs: Optional arguments that ``request`` takes.

        Returns:
            requests.Response
        """
        kwargs.update({k: v for k, v in self.kwargs.items() if k not in kwargs.keys()})

        if self.session:
            return self.session.head(url=url, params=params, **kwargs)
        return requests.head(url=url, params=params, **kwargs)

    def post(
        self,
        url: str,
        data: Union[Dict, ByteString, List[Tuple], IO, str] = None,
        json: Dict = None,
        **kwargs,
    ) -> requests.Response:
        """
        Sends a POST request either using the provided session or the
        single session.

        Args:
            url: URL for the new :class:`Request` object.
            data: Dictionary, list of tuples, bytes, or file-like object to
                send in the body of the :class:`Request`.
            json: A JSON serializable Python object to send in the
                body of the :class:`Request`.
            **kwargs: Optional arguments that ``request`` takes.

        Returns:

        """
        kwargs.update({k: v for k, v in self.kwargs.items() if k not in kwargs.keys()})

        if self.session:
            return self.session.post(url=url, data=data, json=json, **kwargs)
        return requests.post(url=url, data=data, json=json, **kwargs)

    def put(
        self,
        url: str,
        data: Union[Dict, ByteString, List[Tuple], IO, str] = None,
        json: Dict = None,
        **kwargs,
    ) -> requests.Response:
        """
        Sends a PUT request either using the provided session or the
        single session.

        Args:
            url: URL for the new :class:`Request` object.
            data (Union[Dict, ByteString, List[Tuple], IO]):
                Dictionary, list of tuples, bytes, or file-like
                object to send in the body of the :class:`Request`.
            json (Dict): A JSON serializable Python object to send in the
                body of the :class:`Request`..
            **kwargs: Optional arguments that ``request`` takes.

        Returns:
            request.Response
        """
        kwargs.update({k: v for k, v in self.kwargs.items() if k not in kwargs.keys()})

        if self.session:
            return self.session.put(url=url, data=data, json=json, **kwargs)
        return requests.put(url=url, data=data, json=json, **kwargs)

    def patch(
        self,
        url: str,
        data: Union[Dict, ByteString, List[Tuple], IO, str] = None,
        json: Dict = None,
        **kwargs,
    ) -> requests.Response:
        """
        Sends a PATCH request either using the provided session or the
        single session.

        Args:
            url: URL for the new :class:`Request` object.
            data (Union[Dict, ByteString, List[Tuple], IO]):
                Dictionary, list of tuples, bytes, or file-like
                object to send in the body of the :class:`Request`.
            json (Dict): A JSON serializable Python object to send in the
                body of the :class:`Request`..
            **kwargs: Optional arguments that ``request`` takes.

        Returns:
            request.Response
        """
        kwargs.update({k: v for k, v in self.kwargs.items() if k not in kwargs.keys()})

        if self.session:
            return self.session.patch(url=url, data=data, json=json, **kwargs)
        return requests.patch(url=url, data=data, json=json, **kwargs)

    def delete(self, url: str, **kwargs) -> requests.Response:
        """
        Sends a DELETE request either using the provided session or the
        single session.

        Args:
            url (str): URL for the new :class:`Request` object.
            **kwargs: Optional arguments that ``request`` takes.

        Returns:
            request.Response
        """
        kwargs.update({k: v for k, v in self.kwargs.items() if k not in kwargs.keys()})

        if self.session:
            return self.session.delete(url=url, **kwargs)
        return requests.delete(url=url, **kwargs)

    def log_error(self, err: requests.RequestException, msg: str = None) -> None:
        """
        Outputs the error messages from the client request function. If
        additional information is available in the server response this will
        be forwarded to the logging output.

        Note:
            The user is responsible to setup the logging system

        Args:
            err: Request Error
            msg: error message from calling function

        Returns:
            None
        """
        if err.response is not None:
            if err.response.text and msg:
                self.logger.error("%s \n Reason: %s", msg, err.response.text)
            elif err.response.text and not msg:
                self.logger.error("%s", err.response.text)
        elif not err.response and msg:
            self.logger.error("%s \n Reason: %s", msg, err)
        else:
            self.logger.error(err)

    def close(self) -> None:
        """
        Close http session
        Returns:
            None
        """
        if self.session and not self._external_session:
            self.session.close()



"""
Test module for configuration functions
"""

import os
import unittest
from filip.config import Settings
import json
from pydantic import AnyHttpUrl


class TestSettings(unittest.TestCase):
    """
    Test case for loading settings
    """

    def setUp(self) -> None:

        # Test if the testcase was run directly or over in a global test-run.
        # Match the needed path to the config file in both cases
        if os.getcwd().split("\\")[-1] == "tests":
            self.settings_parsing = Settings(_env_file="test_config.env")
        else:
            self.settings_parsing = Settings(_env_file="./tests/test_config.env")

        for key, value in json.loads(self.settings_parsing.model_dump_json()).items():
            os.environ[key] = value
        self.settings_dotenv = Settings()

    def test_load_dotenv(self):
        """
        Tests loading form dotenv file

        Returns:
            None
        """
        self.assertEqual(
            str(self.settings_parsing.IOTA_URL), str(AnyHttpUrl("http://myHost:4041/"))
        )
        self.assertEqual(
            str(self.settings_parsing.CB_URL), str(AnyHttpUrl("http://myHost:1026/"))
        )
        self.assertEqual(
            str(self.settings_parsing.QL_URL), str(AnyHttpUrl("http://myHost:8668/"))
        )

    def test_example_dotenv(self):
        """
        Tests to parse settings

        Returns:
            None
        """
        self.assertEqual(self.settings_parsing, self.settings_dotenv)

    def tearDown(self) -> None:
        for k in self.settings_parsing.model_dump().keys():
            del os.environ[k]
